#!/usr/bin/env python3

import re
import numpy as np
import pandas as pd
import os
import json

import argparse

parser = argparse.ArgumentParser()

parser.add_argument("-dir", "--subdirectory", help="enter subdirectory name")
parser.add_argument("--suffix", help="suffix to add to the name of the files")

params = parser.parse_args()

subdir = params.subdirectory
suffix = params.suffix

if suffix is None:
    suffix = '_concat'

os.chdir(subdir)

cwd = os.getcwd()
units_dir = 'units'

# get directories
with open(os.path.join(units_dir, 'units.txt'), "r") as txt_file:
    fms = txt_file.read().split('\n')

dirs = [fm for fm in fms if fm]
print(dirs)

# combine dataframes
df_loc = []
df_acc = []
df_ri_info = []
df_ri_scope = []

# add each fm files
for fm_next in dirs:
    loc_filepath = os.path.join(units_dir, fm_next, 'location.csv')
    df_loc_tmp = pd.read_csv(loc_filepath, dtype=str, keep_default_na=False)
    df_loc_tmp['FlexiLocUnit'] = fm_next
    df_loc.append(df_loc_tmp)

    acc_filepath = os.path.join(units_dir, fm_next, 'account.csv')
    df_acc_tmp = pd.read_csv(acc_filepath, dtype=str, keep_default_na=False)
    df_acc_tmp['FlexiAccUnit'] = fm_next
    # concat files
    df_acc.append(df_acc_tmp)

    ri_info_filepath = os.path.join(units_dir, fm_next, 'ri_info.csv')
    ri_scope_filepath = os.path.join(units_dir, fm_next, 'ri_scope.csv')
    if os.path.isfile(ri_info_filepath) and os.path.isfile(ri_scope_filepath):
        df_ri_info_tmp = pd.read_csv(ri_info_filepath, dtype=str, keep_default_na=False)
        df_ri_scope_tmp = pd.read_csv(ri_scope_filepath, dtype=str, keep_default_na=False)
        df_ri_info.append(df_ri_info_tmp)
        df_ri_scope.append(df_ri_scope_tmp)

pd.concat(df_loc).to_csv(f'location{suffix}.csv', index=False)
pd.concat(df_acc).to_csv(f'account{suffix}.csv', index=False)

if df_ri_info and df_ri_scope:
    pd.concat(df_ri_info).to_csv(f'ri_info{suffix}.csv', index=False)
    pd.concat(df_ri_scope).to_csv(f'ri_scope{suffix}.csv', index=False)



# Note that only the currently used fields are shown unless show_all is set to True.
import os
import pandas as pd
import anytree
from anytree.search import find
from anytree.exporter import DotExporter
import collections

PolicyTuple = collections.namedtuple('PolicyTuple', 'layer_id agg_id calc_rules')
CalcRuleTuple = collections.namedtuple('CalcRuleTuple', 'policytc_id calcrule_id is_step trig_start trig_end')


def load_df(path, required_file=None):
    if path:
        return pd.read_csv(path)
    else:
        if required_file:
            raise FileNotFoundError(f"Required File does not exist: {required_file}")
        else:
            return None


def create_fm_tree(fm_programme_df, fm_policytc_df, fm_profile_df, fm_summary_df):
    missing_node_link = False

    def get_policy_tc(agg_id, level_id):
        policytc = fm_policytc_df.loc[
            (fm_policytc_df['agg_id'] == agg_id) & (fm_policytc_df['level_id'] == level_id)
        ]

        policy_list = []
        for _, policy in policytc.iterrows():

            # Find calc_rule
            profile = fm_profile_df.loc[fm_profile_df.policytc_id == policy.policytc_id]
            calc_rules = []
            for _, step in profile.iterrows():
                trig_start = step.trigger_start if hasattr(step, 'trigger_start') else 0
                trig_end = step.trigger_end if hasattr(step, 'trigger_end') else 0
                is_step_rule = (trig_end > 0 or trig_start > 0)

                calc_rules.append(CalcRuleTuple(
                    policytc_id=int(policy.policytc_id),
                    calcrule_id=int(step.calcrule_id),
                    is_step=is_step_rule,
                    trig_start=trig_start,
                    trig_end=trig_end,
                ))

            policy_list.append(
                PolicyTuple(
                    layer_id=int(policy.layer_id),
                    agg_id=int(policy.agg_id),
                    calc_rules=calc_rules,
                )
            )
        return len(policytc), policy_list

    level_ids = sorted(list(fm_programme_df.level_id.unique()), reverse=True)
    root = anytree.Node('Insured Loss', agg_id=1, level_id=max(level_ids) + 1, policy_tc=None)

    for level in level_ids:
        agg_id_idxs = list(fm_programme_df[fm_programme_df.level_id == level].drop_duplicates(subset=['level_id', 'to_agg_id'], keep="first").index)

        for node_idx in agg_id_idxs:
            node_info = fm_programme_df.iloc[node_idx]
            layer_max, policy_list = get_policy_tc(node_info.to_agg_id, node_info.level_id)

            # Set parent node as root or find based on level/agg ids
            if level == max(level_ids):
                parent_node = root
            else:
                try:
                    matched_id = fm_programme_df.loc[(fm_programme_df.level_id == level + 1) &
                                                     (fm_programme_df.from_agg_id == node_info.to_agg_id)].to_agg_id.item()
                    parent_node = find(root, filter_=lambda node: node.level_id == level + 1 and node.agg_id == matched_id)
                except ValueError:
                    missing_node_link = True
                    print('Missing node link: agg_id={}, level_id={}'.format(node_info.to_agg_id, level + 1))

            # Set node names based on attrs in FM files
            if level >= 3:
                node_name = "policy term {} \nlevel: {}".format(
                    node_info.to_agg_id,
                    node_info.level_id
                )
            elif level == 2:
                node_name = "loc term {} ".format(node_info.to_agg_id)
            else:
                node_name = "cov term {}".format(node_info.to_agg_id)

            for policy in policy_list:
                node_name += "\n\nlayer_id: {}".format(policy.layer_id)
                for rule in policy.calc_rules:
                    if rule.is_step:
                        node_name += "\n   policytc_id {}: step_rule:{}, start:{} end:{}".format(
                            rule.policytc_id,
                            rule.calcrule_id,
                            rule.trig_start,
                            rule.trig_end
                        )
                    else:
                        node_name += "\npolicytc_id: {} \ncalc_rule: {}".format(
                            rule.policytc_id,
                            rule.calcrule_id,
                        )

            # Create Node in FM tree
            node = anytree.Node(
                node_name,
                agg_id=node_info.to_agg_id,
                level_id=level,
                parent=parent_node,
                layer_max=layer_max,
                policy_tc=policy_list,
            )

    # Add item level data
    item_agg_idx = list(fm_summary_df[['agg_id']].drop_duplicates().index)
    for item in item_agg_idx:
        item_info = fm_summary_df.iloc[item]
        matched_id = fm_programme_df.loc[(fm_programme_df.level_id == 1) & (fm_programme_df.from_agg_id == item_info.agg_id)].to_agg_id.item()
        parent_node = find(root, filter_=lambda node: node.level_id == 1 and node.agg_id == matched_id)

        node_name = "\n".join([
            "item {}\n".format(int(item_info.agg_id)),
            "locnumber: {}".format(item_info.locnumber),
            "accnumber: {}".format(item_info.accnumber),
            "polnumber: {}".format(item_info.polnumber),
            "portnumber: {}".format(item_info.portnumber),
            "cov_type: {}".format(item_info.coverage_type_id),
            "peril_id: {}".format(item_info.peril_id),
            "tiv: {}".format(item_info.tiv),
        ])

        node = anytree.Node(
            node_name,
            agg_id=item_info.agg_id,
            level_id=0,
            parent=parent_node,
            locnumber=item_info.locnumber,
            accnumber=item_info.accnumber,
            polnumber=item_info.polnumber,
            portnumber=item_info.polnumber,
            tiv=item_info.tiv,
            coverage_id=item_info.coverage_id,
            coverage_type=item_info.coverage_type_id,
            peril_id=item_info.peril_id,
        )
    return root, missing_node_link


def render_fm_tree(root_node, filename='tree.png'):

    # Function to format nodes in FM tree
    def format_box(node):
        # https://graphviz.org/doc/info/shapes.html
        if node.level_id == 0:
            # Item Level Node
            return "fixedsize=false, shape=rect, fillcolor=lightgrey, style=filled"
        else:
            if not node.policy_tc:
                # Error? missing policy_tc entry for this Node
                return "fixedsize=false, shape=ellipse, fillcolor=pink, style=filled"

            elif len(node.policy_tc) > 1:
                # Node with multiple layers
                return "fixedsize=false, shape=rect, fillcolor=orange, style=filled"
            else:
                # Cov or loc nodes
                return "fixedsize=false, shape=ellipse, fillcolor=lightblue, style=filled"

    # Function to add weighted 'by layer number' edges
    def layered_edge(node, child):
        # https://anytree.readthedocs.io/en/latest/tricks/weightededges.html
        if hasattr(child, 'layer_max'):
            if child.layer_max > 1:
                return 'dir=back, style=bold, label=" {} Layers"'.format(child.layer_max)
        return "dir=back"

    # Render tree to png
    dot_data = DotExporter(
        root_node,
        edgeattrfunc=layered_edge,
        nodeattrfunc=format_box)
    dot_data.to_picture(filename)



__all__ = [
    'GenerateKeys',
    'GenerateKeysDeterministic'
]

import os
from packaging import version

from ..base import ComputationStep
from ...lookup.factory import KeyServerFactory
from ...utils.exceptions import OasisException

from ...utils.inputs import str2bool
from ...utils.data import get_utctimestamp, get_exposure_data, analysis_settings_loader, model_settings_loader


class KeyComputationStep(ComputationStep):
    def get_exposure_data_config(self):
        return {
            'location': self.oed_location_csv,
            'oed_schema_info': self.oed_schema_info if self.oed_schema_info is not None else self.settings.get('oed_version', None),
            'check_oed': self.check_oed,
            'use_field': True,
            'backend_dtype': self.oed_backend_dtype,
        }


class GenerateKeys(KeyComputationStep):
    """
    Generates keys from a model lookup, and write Oasis keys and keys error files.

    The model lookup, which is normally independently implemented by the model
    supplier, should generate keys as dicts with the following format
    ::

        {
            "id": <loc. ID>,
            "peril_id": <OED sub-peril ID>,
            "coverage_type": <OED coverage type ID>,
            "area_peril_id": <area peril ID>,
            "vulnerability_id": <vulnerability ID>,
            "message": <loc. lookup status message>,
            "status": <loc. lookup status flag indicating success, failure or no-match>
        }

    The keys generation command can generate these dicts, and write them to
    file. It can also be used to write these to an Oasis keys file (which is a
    requirement for model execution), which has the following format.::

        LocID,PerilID,CoverageTypeID,AreaPerilID,VulnerabilityID
        ..
        ..
    This file only lists the locations for which there has been a successful
    lookup. The keys errors file lists all the locations with failing or
    non-matching lookups and has the following format::

        LocID,PerilID,CoverageTypeID,Message
        ..
        ..
    """
    settings_params = [{'name': 'lookup_complex_config_json', 'loader': analysis_settings_loader, 'user_role': 'user'},
                       {'name': 'model_settings_json', 'loader': model_settings_loader}]

    step_params = [
        {'name': 'oasis_files_dir', 'is_path': True, 'pre_exist': False,
         'help': 'Path to the directory in which to generate the Oasis files'},
        {'name': 'oed_location_csv', 'flag': '-x', 'is_path': True, 'pre_exist': True, 'help': 'Source location CSV file path'},
        {'name': 'oed_schema_info', 'help': 'Takes a version of OED schema to use in the form "v1.2.3" or a path to an OED schema json'},
        {'name': 'check_oed', 'type': str2bool, 'const': True, 'nargs': '?', 'default': True, 'help': 'if True check input oed files'},
        {'name': 'keys_data_path', 'flag': '-k', 'is_path': True, 'pre_exist': False, 'help': 'Generated keys CSV output path'},
        {'name': 'keys_errors_path', 'flag': '-e', 'is_path': True, 'pre_exist': False, 'help': 'Generated keys errors CSV output path'},
        {'name': 'keys_format', 'help': 'Keys files output format', 'choices': ['oasis', 'json', 'parquet'], 'default': 'parquet'},
        {'name': 'lookup_config_json', 'flag': '-g', 'is_path': True, 'pre_exist': False, 'help': 'Lookup config JSON file path'},
        {'name': 'lookup_data_dir', 'is_path': True, 'pre_exist': True, 'help': 'Model lookup/keys data directory path'},
        {'name': 'lookup_module_path', 'flag': '-l', 'is_path': True, 'pre_exist': False, 'help': 'Model lookup module path'},
        {'name': 'lookup_complex_config_json', 'is_path': True, 'pre_exist': False, 'help': 'Complex lookup config JSON file path'},
        {'name': 'lookup_num_processes', 'type': int, 'default': -1, 'help': 'Number of workers in multiprocess pools'},
        {'name': 'lookup_num_chunks', 'type': int, 'default': -1, 'help': 'Number of chunks to split the location file into for multiprocessing'},
        {'name': 'model_version_csv', 'flag': '-v', 'is_path': True, 'pre_exist': False, 'help': 'Model version CSV file path'},
        {'name': 'model_settings_json', 'flag': '-M', 'is_path': True, 'pre_exist': True, 'help': 'Model settings JSON file path'},
        {'name': 'user_data_dir', 'flag': '-D', 'is_path': True, 'pre_exist': False,
         'help': 'Directory containing additional model data files which varies between analysis runs'},
        {'name': 'lookup_multiprocessing', 'type': str2bool, 'const': True, 'nargs': '?', 'default': True,
         'help': 'Flag to enable/disable lookup multiprocessing'},
        {'name': 'disable_oed_version_update', 'type': str2bool, 'const': True, 'nargs': '?', 'default': False,
         'help': 'Flag to enable/disable conversion to latest compatible OED version. Must be present in model settings.'},
        {'name': 'oed_backend_dtype', 'type': str, 'default': 'pd_dtype',
         'help': "define what type dtype the oed column will be (pd_dtype or pa_dtype)"},

        # Manager only options
        {'name': 'verbose', 'default': False},
    ]

    def _get_output_dir(self):
        if self.keys_data_path:
            return os.path.dirname(self.keys_data_path)
        elif self.oasis_files_dir:
            return self.oasis_files_dir
        else:
            utcnow = get_utctimestamp(fmt='%Y%m%d%H%M%S')
            return os.path.join(os.getcwd(), 'runs', 'keys-{}'.format(utcnow))

    def run(self):
        if not (self.lookup_config_json or (self.lookup_data_dir and self.model_version_csv and self.lookup_module_path)):
            raise OasisException(
                'No pre-generated keys file provided, and no lookup assets '
                'provided to generate a keys file - if you do not have a '
                'pre-generated keys file then lookup assets must be provided - '
                'for a built-in lookup the lookup config. JSON file path must '
                'be provided, or for custom lookups the keys data path + model '
                'version file path + lookup package path must be provided'
            )
        self.oasis_files_dir = self._get_output_dir()
        output_type = 'csv' if self.keys_format.lower() == 'oasis' else self.keys_format.lower()

        exposure_data = get_exposure_data(self, add_internal_col=True)

        if not self.disable_oed_version_update:
            supported_versions = self.settings.get('data_settings', {}).get('supported_oed_versions', None)
            if supported_versions:
                if isinstance(supported_versions, str):
                    self.logger.info(f"Converting to OED version {supported_versions}")
                    exposure_data.to_version(supported_versions)
                elif isinstance(supported_versions, list) and supported_versions:
                    # If 'supported_oed_versions' is a list and is not empty
                    # Sort the versions in descending order
                    supported_versions = sorted(supported_versions, key=version.parse, reverse=True)
                    self.logger.info(f"Converting to OED version {supported_versions[0]}")
                    exposure_data.to_version(supported_versions[0])
                else:
                    # If 'supported_oed_versions' is neither a string nor a non-empty list
                    self.logger.warning("Invalid OED version information in model settings.")
            else:
                # If 'supported_oed_versions' is missing or empty
                self.logger.debug("No OED version information in model settings.")

        keys_fp = self.keys_data_path or os.path.join(self.oasis_files_dir, f'keys.{output_type}')
        keys_errors_fp = self.keys_errors_path or os.path.join(self.oasis_files_dir, f'keys-errors.{output_type}')
        os.makedirs(os.path.dirname(keys_fp), exist_ok=True)
        os.makedirs(os.path.dirname(keys_errors_fp), exist_ok=True)

        keys_success_msg = True if self.lookup_complex_config_json else False

        model_info, key_server = KeyServerFactory.create(
            lookup_config_fp=self.lookup_config_json,
            model_keys_data_path=self.lookup_data_dir,
            model_version_file_path=self.model_version_csv,
            lookup_module_path=self.lookup_module_path,
            complex_lookup_config_fp=self.lookup_complex_config_json,
            user_data_dir=self.user_data_dir,
            output_directory=self.oasis_files_dir
        )

        res = key_server.generate_key_files(
            location_df=exposure_data.get_subject_at_risk_source().dataframe,
            successes_fp=keys_fp,
            errors_fp=keys_errors_fp,
            output_format=self.keys_format,
            keys_success_msg=keys_success_msg,
            multiproc_enabled=self.lookup_multiprocessing,
            multiproc_num_cores=self.lookup_num_processes,
            multiproc_num_partitions=self.lookup_num_chunks,
        )

        self.logger.debug(f"key generated used model {model_info}")
        self.logger.info('\nKeys successful: {} generated with {} items'.format(res[0], res[1]))
        if len(res) == 4:
            self.logger.info('Keys errors: {} generated with {} items'.format(res[2], res[3]))
        return res


class GenerateKeysDeterministic(KeyComputationStep):
    step_params = [
        {'name': 'oasis_files_dir', 'is_path': True, 'pre_exist': False,
         'help': 'Path to the directory in which to generate the Oasis files'},
        {'name': 'oed_location_csv', 'flag': '-x', 'is_path': True, 'pre_exist': True, 'help': 'Source location CSV file path'},
        {'name': 'oed_schema_info', 'help': 'Takes a version of OED schema to use in the form "v1.2.3" or a path to an OED schema json'},
        {'name': 'check_oed', 'type': str2bool, 'const': True, 'nargs': '?', 'default': True, 'help': 'if True check input oed files'},
        {'name': 'keys_data_path', 'flag': '-k', 'is_path': True, 'pre_exist': False, 'help': 'Generated keys CSV output path'},
        {'name': 'keys_format', 'help': 'Keys files output format',
         'choices': ['oasis', 'json', 'parquet'], 'default': 'parquet'},
        {'name': 'supported_oed_coverage_types', 'type': int, 'nargs': '+', 'help': 'Select List of supported coverage_types [1, .. ,15]'},
        {'name': 'model_perils_covered', 'nargs': '+', 'default': ['AA1'],
         'help': 'List of peril covered by the model'},
        {'name': 'oed_backend_dtype', 'type': str, 'default': 'pd_dtype',
         'help': "define what type dtype the oed column will be (pd_dtype or pa_dtype)"},
    ]

    def _get_output_dir(self):
        if self.keys_data_path:
            return os.path.dirname(self.keys_data_path)
        elif self.oasis_files_dir:
            return self.oasis_files_dir
        else:
            utcnow = get_utctimestamp(fmt='%Y%m%d%H%M%S')
            return os.path.join(os.getcwd(), 'runs', 'keys-{}'.format(utcnow))

    def run(self):
        self.oasis_files_dir = self._get_output_dir()
        output_type = 'csv' if self.keys_format.lower() == 'oasis' else self.keys_format.lower()
        keys_fp = self.keys_data_path or os.path.join(self.oasis_files_dir, f'keys.{output_type}')
        os.makedirs(os.path.dirname(keys_fp), exist_ok=True)

        exposure_data = get_exposure_data(self, add_internal_col=True)

        if self.supported_oed_coverage_types is None:
            coverage_values = exposure_data.oed_schema.schema['CoverageValues']
            cob_coverage = list(
                coverage_info['CoverageID'] for coverage_info in coverage_values.values()
                if not coverage_info['SubCoverages']
                and coverage_info['Type'] == exposure_data.class_of_business_info['name'])
            self.supported_oed_coverage_types = cob_coverage

        config = {'builtin_lookup_type': 'peril_covered_deterministic',
                  'model': {"supplier_id": "OasisLMF",
                            "model_id": "Deterministic",
                            "model_version": "1"},
                  'supported_oed_coverage_types': self.supported_oed_coverage_types,
                  'model_perils_covered': self.model_perils_covered}

        model_info, lookup = KeyServerFactory.create(
            lookup_config=config,
            output_directory=self.oasis_files_dir
        )

        return lookup.generate_key_files(
            location_df=exposure_data.get_subject_at_risk_source().dataframe,
            successes_fp=keys_fp,
            output_format=self.keys_format,
        )





from glob import glob
from importlib import import_module

cli_module_names = (m.split('/')[-1].split('.')[0] for m in glob("oasislmf/cli/*") if m.endswith('.py'))

for cli_module_name in cli_module_names:

    main_class = getattr(
        import_module(f"oasislmf.cli.{cli_module_name}"),
        f'{cli_module_name.capitalize()}Cmd',
        None
    )

    print(main_class, cli_module_name)
    if main_class:
        sub_commands = main_class.sub_commands
        print(cli_module_name, sub_commands)
        for command, class_name in sub_commands.items():

            tmpl = f"""
            ``{cli_module_name}``
            {'-'*len(cli_module_name)}
            .. automodule:: oasislmf.cli.{cli_module_name}.{class_name}
            :noindex:
            """

            print(tmpl)



import numba as nb
import numpy as np

from oasislmf.pytools.common.data import areaperil_int, oasis_float, oasis_int, item_adjustment_dtype, item_id, NAME_DTYPE_SLICE

# gul stream type
# probably need to set this dynamically depending on the stream type
gul_header = np.int32(1 | 2 << 24).tobytes()

# define the damage_bin_dict damage_types
DAMAGE_TYPE_DEFAULT = 0
DAMAGE_TYPE_RELATIVE = 1
DAMAGE_TYPE_ABSOLUTE = 2
DAMAGE_TYPE_DURATION = 3
VALID_DAMAGE_TYPE = {DAMAGE_TYPE_DEFAULT, DAMAGE_TYPE_RELATIVE, DAMAGE_TYPE_ABSOLUTE, DAMAGE_TYPE_DURATION}

ItemAdjustment = nb.from_dtype(item_adjustment_dtype)

items_data_type = nb.from_dtype(np.dtype([item_id[NAME_DTYPE_SLICE],
                                          ('damagecdf_i', oasis_int),
                                          ('rng_index', oasis_int)
                                          ]))

items_MC_data_type = nb.from_dtype(np.dtype([item_id[NAME_DTYPE_SLICE],
                                             ('item_idx', oasis_int),
                                             ('haz_arr_i', oasis_int),
                                             ('rng_index', oasis_int),
                                             ('hazard_rng_index', oasis_int),
                                             ('intensity_adjustment', oasis_int),
                                             ('return_period', oasis_int)
                                             ]))

VulnCdfLookup = nb.from_dtype(np.dtype([('start', oasis_int),
                                        ('length', oasis_int)]))

coverage_type = nb.from_dtype(np.dtype([('tiv', np.float64),
                                        ('max_items', np.int32),
                                        ('start_items', np.int32),
                                        ('cur_items', np.int32)
                                        ]))

NP_BASE_ARRAY_SIZE = 8


ITEM_MAP_KEY_TYPE = nb.types.Tuple((nb.from_dtype(areaperil_int), nb.types.int32))
# ITEM_MAP_VALUE_TYPE = nb.types.UniTuple(nb.types.int32, 1)
ITEM_MAP_VALUE_TYPE = nb.types.int64
AREAPERIL_TO_EFF_VULN_KEY_TYPE = nb.types.Tuple((nb.from_dtype(areaperil_int), nb.types.int64))
AREAPERIL_TO_EFF_VULN_VALUE_TYPE = nb.types.UniTuple(nb.types.int32, 2)

# compute the relative size of oasis_float and areaperil_int vs int32
oasis_float_to_int32_size = oasis_float.itemsize // np.int32().itemsize
areaperil_int_to_int32_size = areaperil_int.itemsize // np.int32().itemsize

haz_arr_type = nb.from_dtype(np.dtype([('probability', oasis_float),
                                       ('intensity_bin_id', np.int32),
                                       ('intensity', np.int32)]))

ProbMean = nb.from_dtype(np.dtype([('prob_to', oasis_float),
                                   ('bin_mean', oasis_float)
                                   ]))
ProbMean_size = ProbMean.size

damagecdfrec_stream = nb.from_dtype(np.dtype([('event_id', np.int32),
                                              ('areaperil_id', areaperil_int),
                                              ('vulnerability_id', np.int32)
                                              ]))

damagecdfrec = nb.from_dtype(np.dtype([('areaperil_id', areaperil_int),
                                       ('vulnerability_id', np.int32)
                                       ]))


gulSampleslevelHeader = nb.from_dtype(np.dtype([('event_id', 'i4'),
                                                ('item_id', 'i4'),
                                                ]))
gulSampleslevelHeader_size = gulSampleslevelHeader.size

gulSampleslevelRec = nb.from_dtype(np.dtype([('sidx', 'i4'),
                                             ('loss', oasis_float),
                                             ]))
gulSampleslevelRec_size = gulSampleslevelRec.size


Keys = {'LocID': np.int32,
        'PerilID': 'category',
        'CoverageTypeID': np.int32,
        'AreaPerilID': areaperil_int,
        'VulnerabilityID': np.int32}


NormInversionParameters = nb.from_dtype(np.dtype([('x_min', np.float64),
                                                  ('x_max', np.float64),
                                                  ('N', np.int32),
                                                  ('cdf_min', np.float64),
                                                  ('cdf_max', np.float64),
                                                  ('inv_factor', np.float64),
                                                  ('norm_factor', np.float64),
                                                  ]))

gulmc_compute_info_type = nb.from_dtype(np.dtype([
    ('event_id', oasis_int),
    ('cursor', oasis_int),
    ('coverage_i', oasis_int),  # last_processed_coverage_ids_idx
    ('coverage_n', oasis_int),
    ('next_cached_vuln_cdf_i', oasis_int),
    ('max_bytes_per_item', oasis_int),
    ('Ndamage_bins_max', oasis_int),
    ('loss_threshold', oasis_float),
    ('alloc_rule', np.int8),
    ('do_correlation', np.int8),
    ('do_haz_correlation', np.int8),
    ('effective_damageability', np.int8),
    ('debug', np.int8),
]))



import os
import sys
import numba as nb
import numpy as np
import pandas as pd


oasis_int = np.dtype(os.environ.get('OASIS_INT', 'i4'))
nb_oasis_int = nb.from_dtype(oasis_int)
oasis_int_size = oasis_int.itemsize

oasis_float = np.dtype(os.environ.get('OASIS_FLOAT', 'f4'))
nb_oasis_float = nb.from_dtype(oasis_float)
oasis_float_size = oasis_float.itemsize

areaperil_int = np.dtype(os.environ.get('OASIS_AREAPERIL_TYPE', 'u4'))
nb_areaperil_int = nb.from_dtype(areaperil_int)
areaperil_int_size = areaperil_int.itemsize

null_index = oasis_int.type(-1)

# A default buffer size for nd arrays to be initialised to
DEFAULT_BUFFER_SIZE = 1_000_000

# Mean type numbers for outputs (SampleType)
MEAN_TYPE_ANALYTICAL = 1
MEAN_TYPE_SAMPLE = 2


def generate_output_metadata(output):
    """Generates *_header, *_dtype and *_fmt items given a list of tuples describing some output description
    output description has type List(Tuple({name: str}, {type: Any}, {format: str}))
    Args:
        output_map (list(tuple(str, Any, str))): Dictionary mapping string name to  {output description}_output list
    Returns:
        result (tuple(list[str], np.dtype, str)): Tuple containing the generated *_header list, *_dtype np.dtype, *_fmt csv format string
    """
    headers = [c[0] for c in output]
    dtype = np.dtype([(c[0], c[1]) for c in output])
    fmt = ','.join([c[2] for c in output])
    result = (headers, dtype, fmt)
    return result


# single type definition index
DTYPE_IDX = 1
NAME_DTYPE_SLICE = slice(2)

# single type definition (alphabetical order)
agg_id = ("agg_id", 'i4', "%d")
aggregate_vulnerability_id = ("aggregate_vulnerability_id", 'i4', "%d")
amplification_id = ("amplification_id", 'i4', "%d")
areaperil_id = ("areaperil_id", areaperil_int, "%d")
attachment1 = ("attachment1", oasis_float, "%f")
bin_from = ("bin_from", oasis_float, "%f")
bin_index = ("bin_index", 'i4', "%d")
bin_mean = ("bin_mean", 'f4', "%f")
bin_to = ("bin_to", oasis_float, "%f")
calcrule_id = ("calcrule_id", 'i4', "%d")
coverage_id = ("coverage_id", 'u4', "%u")
damage_bin_id = ("damage_bin_id", 'i4', "%d")
damage_correlation_value = ("damage_correlation_value", oasis_float, "%f")
damage_type = ("damage_type", 'i4', "%d")
deductible1 = ("deductible1", oasis_float, "%f")
deductible2 = ("deductible2", oasis_float, "%f")
deductible3 = ("deductible3", oasis_float, "%f")
event_id = ("event_id", 'i4', "%d")
factor = ("factor", 'f4', "%.2f")  # Amplification factor
from_agg_id = ("from_agg_id", 'i4', "%d")
group_id = ("group_id", 'u4', "%u")
hazard_correlation_value = ("hazard_correlation_value", oasis_float, "%f")
hazard_group_id = ("hazard_group_id", 'i4', "%d")
intensity_adjustment = ('intensity_adjustment', 'i4', "%d")
intensity_bin_id = ('intensity_bin_id', 'i4', "%d")
interpolation = ("interpolation", oasis_float, "%f")
item_id = ("item_id", 'i4', "%d")
item_return_period = ('return_period', 'i4', "%d")
layer_id = ("layer_id", 'i4', "%d")
level_id = ("level_id", 'i4', "%d")
limit1 = ("limit1", oasis_float, "%f")
limit2 = ("limit2", oasis_float, "%f")
loss = ("loss", oasis_float, "%.2f")
model_data_len = ("model_data_len", 'u4', "%u")
occ_date_id = ("occ_date_id", 'i4', "%d")
occ_date_id_granular = ("occ_date_id", 'i8', "%d")
output_id = ("output_id", 'i4', "%d")
output = ("output", output_id[1], output_id[2])
payout_end = ("payout_end", oasis_float, "%f")
payout_start = ("payout_start", oasis_float, "%f")
peril_correlation_group = ("peril_correlation_group", 'i4', "%d")
period_no = ("period_no", 'i4', "%d")
period_weighting = ("weighting", 'f8', "%0.9lf")
prob_to = ("prob_to", 'f4', "%f")
probability = ('probability', oasis_float, "%.6f")
profile_id = ("profile_id", 'i4', "%d")
quantile = ("quantile", 'f4', "%f")
quantile_fractional_part = ('fractional_part', oasis_float, "%f")
quantile_integer_part = ('integer_part', oasis_int, "%d")
random_no = ("random_no", 'f4', "%f")
return_period = ("return_period", 'i4', "%d")
scale1 = ("scale1", oasis_float, "%f")
scale2 = ("scale2", oasis_float, "%f")
section_id = ("scale2", oasis_int, "%d")
share1 = ("share1", oasis_float, "%f")
share2 = ("share2", oasis_float, "%f")
share3 = ("share3", oasis_float, "%f")
sidx = ("sidx", 'i4', "%d")
step_id = ("step_id", 'i4', "%d")
summary_id = ("summary_id", 'i4', "%d")
summaryset_id = ("summaryset_id", 'i4', "%d")
tiv = ("tiv", oasis_float, "%f")
to_agg_id = ("to_agg_id", 'i4', "%d")
trigger_end = ("trigger_end", oasis_float, "%f")
trigger_start = ("trigger_start", oasis_float, "%f")
vulnerability_id = ("vulnerability_id", 'i4', "%d")
vulnerability_weight = ("weight", oasis_float, "%f")


# Types
aggregatevulnerability_output = [
    aggregate_vulnerability_id,
    vulnerability_id,
]
aggregatevulnerability_headers, aggregatevulnerability_dtype, aggregatevulnerability_fmt = generate_output_metadata(aggregatevulnerability_output)

amplifications_output = [
    item_id,
    amplification_id,
]
amplifications_headers, amplifications_dtype, amplifications_fmt = generate_output_metadata(amplifications_output)

cdf_output = [
    event_id,
    areaperil_id,
    vulnerability_id,
    bin_index,
    prob_to,
    bin_mean,
]
cdf_headers, cdf_dtype, cdf_fmt = generate_output_metadata(cdf_output)

complex_items_meta_output = [
    item_id,
    coverage_id,
    group_id,
    model_data_len,
]
complex_items_meta_headers, complex_items_meta_dtype, complex_items_meta_fmt = generate_output_metadata(complex_items_meta_output)

correlations_output = [
    item_id,
    peril_correlation_group,
    damage_correlation_value,
    hazard_group_id,
    hazard_correlation_value,
]
correlations_headers, correlations_dtype, correlations_fmt = generate_output_metadata(correlations_output)

coverages_output = [
    coverage_id,
    tiv,
]
coverages_headers, coverages_dtype, coverages_fmt = generate_output_metadata(coverages_output)

damagebin_output = [
    bin_index,
    bin_from,
    bin_to,
    interpolation,
    damage_type,
]
damagebin_headers, damagebin_dtype, damagebin_fmt = generate_output_metadata(damagebin_output)

eve_output = [
    event_id
]
eve_headers, eve_dtype, eve_fmt = generate_output_metadata(eve_output)

footprint_event_output = [
    event_id,
    areaperil_id,
    intensity_bin_id,
    probability,
]
footprint_event_headers, footprint_event_dtype, footprint_event_fmt = generate_output_metadata(footprint_event_output)

fm_output = [
    event_id,
    output_id,
    sidx,
    loss,
]
fm_headers, fm_dtype, fm_fmt = generate_output_metadata(fm_output)

fm_policytc_output = [
    level_id,
    agg_id,
    layer_id,
    profile_id,
]
fm_policytc_headers, fm_policytc_dtype, fm_policytc_fmt = generate_output_metadata(fm_policytc_output)

fm_profile_output = [
    profile_id,
    calcrule_id,
    deductible1,
    deductible2,
    deductible3,
    attachment1,
    limit1,
    share1,
    share2,
    share3,
]
fm_profile_headers, fm_profile_dtype, fm_profile_fmt = generate_output_metadata(fm_profile_output)

fm_profile_step_output = [
    profile_id,
    calcrule_id,
    deductible1,
    deductible2,
    deductible3,
    attachment1,
    limit1,
    share1,
    share2,
    share3,
    step_id,
    trigger_start,
    trigger_end,
    payout_start,
    payout_end,
    limit2,
    scale1,
    scale2,
]
fm_profile_step_headers, fm_profile_step_dtype, fm_profile_step_fmt = generate_output_metadata(fm_profile_step_output)

fm_programme_output = [
    from_agg_id,
    level_id,
    to_agg_id,
]
fm_programme_headers, fm_programme_dtype, fm_programme_fmt = generate_output_metadata(fm_programme_output)

fm_summary_xref_output = [
    output,
    summary_id,
    summaryset_id
]
fm_summary_xref_headers, fm_summary_xref_dtype, fm_summary_xref_fmt = generate_output_metadata(fm_summary_xref_output)

fm_xref_output = [
    output,
    agg_id,
    layer_id,
]
fm_xref_headers, fm_xref_dtype, fm_xref_fmt = generate_output_metadata(fm_xref_output)

gul_output = [
    event_id,
    item_id,
    sidx,
    loss,
]
gul_headers, gul_dtype, gul_fmt = generate_output_metadata(gul_output)

gul_summary_xref_output = [
    item_id,
    summary_id,
    summaryset_id
]
gul_summary_xref_headers, gul_summary_xref_dtype, gul_summary_xref_fmt = generate_output_metadata(gul_summary_xref_output)

items_output = [
    item_id,
    coverage_id,
    areaperil_id,
    vulnerability_id,
    group_id
]
items_headers, items_dtype, items_fmt = generate_output_metadata(items_output)

item_adjustment = [
    item_id,
    intensity_adjustment,
    item_return_period
]

item_adjustment_headers, item_adjustment_dtype, item_adjustment_fmt = generate_output_metadata(item_adjustment)

lossfactors_output = [
    event_id,
    amplification_id,
    factor,
]
lossfactors_headers, lossfactors_dtype, lossfactors_fmt = generate_output_metadata(lossfactors_output)

occurrence_output = [
    event_id,
    period_no,
    occ_date_id,
]
occurrence_headers, occurrence_dtype, occurrence_fmt = generate_output_metadata(occurrence_output)

occurrence_granular_output = [
    event_id,
    period_no,
    occ_date_id_granular,
]
occurrence_granular_headers, occurrence_granular_dtype, occurrence_granular_fmt = generate_output_metadata(occurrence_granular_output)

periods_output = [
    period_no,
    period_weighting,
]
periods_headers, periods_dtype, periods_fmt = generate_output_metadata(periods_output)

quantile_output = [
    quantile,
]
quantile_headers, quantile_dtype, quantile_fmt = generate_output_metadata(quantile_output)

quantile_interval_output = quantile_output + [
    quantile_integer_part,
    quantile_fractional_part,
]
quantile_interval_headers, quantile_interval_dtype, quantile_interval_fmt = generate_output_metadata(quantile_interval_output)

random_output = [
    random_no,
]
random_headers, random_dtype, random_fmt = generate_output_metadata(random_output)

returnperiods_output = [
    return_period,
]
returnperiods_headers, returnperiods_dtype, returnperiods_fmt = generate_output_metadata(returnperiods_output)

vulnerability_output = [
    vulnerability_id,
    intensity_bin_id,
    damage_bin_id,
    probability,
]
vulnerability_headers, vulnerability_dtype, vulnerability_fmt = generate_output_metadata(vulnerability_output)

vulnerability_weight_output = [
    areaperil_id,
    vulnerability_id,
    vulnerability_weight,
]
vulnerability_weight_headers, vulnerability_weight_dtype, vulnerability_weight_fmt = generate_output_metadata(vulnerability_weight_output)


def load_as_ndarray(dir_path, name, _dtype, must_exist=True, col_map=None):
    """
    load a file as a numpy ndarray
    useful for multi-columns files
    Args:
        dir_path: path to the directory where the binary or csv file is stored
        name: name of the file
        _dtype: np.dtype
        must_exist: raise FileNotFoundError if no file is present
        col_map: name re-mapping to change name of csv columns
    Returns:
        numpy ndarray
    """

    if os.path.isfile(os.path.join(dir_path, name + '.bin')):
        return np.fromfile(os.path.join(dir_path, name + '.bin'), dtype=_dtype)
    elif must_exist or os.path.isfile(os.path.join(dir_path, name + '.csv')):
        # in csv column cam be out of order and have different name,
        # we load with pandas and write each column to the ndarray
        if col_map is None:
            col_map = {}
        with open(os.path.join(dir_path, name + '.csv')) as file_in:
            cvs_dtype = {col_map.get(key, key): col_dtype for key, (col_dtype, _) in _dtype.fields.items()}
            df = pd.read_csv(file_in, delimiter=',', dtype=cvs_dtype, usecols=list(cvs_dtype.keys()))
            res = np.empty(df.shape[0], dtype=_dtype)
            for name in _dtype.names:
                res[name] = df[col_map.get(name, name)]
            return res
    else:
        return np.empty(0, dtype=_dtype)


def load_as_array(dir_path, name, _dtype, must_exist=True):
    """
    load file as a single numpy array,
     useful for files with a binary version with only one type of value where their index correspond to an id.
     For example coverage.bin only contains tiv value for each coverage id
     coverage_id n correspond to index n-1
    Args:
        dir_path: path to the directory where the binary or csv file is stored
        name: name of the file
        _dtype: numpy dtype of the required array
        must_exist: raise FileNotFoundError if no file is present
    Returns:
        numpy array of dtype type
    """
    fp = os.path.join(dir_path, name + '.bin')
    if os.path.isfile(fp):
        return np.fromfile(fp, dtype=_dtype)
    elif must_exist or os.path.isfile(os.path.join(dir_path, name + '.csv')):
        fp = os.path.join(dir_path, name + '.csv')
        with open(fp) as file_in:
            return np.loadtxt(file_in, dtype=_dtype, delimiter=',', skiprows=1, usecols=1)
    else:
        return np.empty(0, dtype=_dtype)


def write_ndarray_to_fmt_csv(output_file, data, headers, row_fmt):
    """Writes a custom dtype array with headers to csv with the provided row_fmt str

    This function is a faster replacement for np.savetxt as it formats each row one at a time before writing to csv.
    We create one large string, and formats all the data at once, and writes all the data at once.

    WARNING: untested with string types in custom data.

    Args:
        output_file (io.TextIOWrapper): CSV file
        data (ndarray[<custom dtype>]): Custom dtype ndarray with column names
        headers (list[str]): Column names for custom ndarray
        row_fmt (str): Format for each row in csv
    """
    if len(headers) != len(row_fmt.split(",")):
        raise RuntimeError(f"ERROR: write_ndarray_to_fmt_csv requires row_fmt ({row_fmt}) and headers ({headers}) to have the same length.")

    # Copy data as np.ravel does not work with custom dtype arrays
    # Default type of np.empty is np.float64.
    data_cpy = np.empty((data.shape[0], len(headers)))
    for i in range(len(headers)):
        data_cpy[:, i] = data[headers[i]]

    # Create one large formatted string
    final_fmt = "\n".join([row_fmt] * data_cpy.shape[0])
    str_data = final_fmt % tuple(np.ravel(data_cpy))

    output_file.write(str_data)
    output_file.write("\n")


float_equal_precision = np.finfo(oasis_float).eps


@nb.njit(cache=True)
def almost_equal(a, b):
    return abs(a - b) < float_equal_precision


def resolve_file(path, mode, stack):
    """Resolve file path to open file or use sys.stdin

    Args:
        path (str | os.PathLike): File path or "-" indicationg standard input/output.
        mode (str): Mode to open file ("r", "rb, "w", "wb").
        stack (ExitStack): Context manager stack used to manage file lifecycle.

    Returns:
        file (IO): A file-like object opened in the specified mode.
    """
    is_read = "r" in mode
    is_binary = "b" in mode

    if str(path) == "-":
        if is_read:
            return sys.stdin.buffer if is_binary else sys.stdin
        else:
            return sys.stdout.buffer if is_binary else sys.stdout
    else:
        return stack.enter_context(open(path, mode))



import io
from IPython.display import display, Javascript, clear_output
import os
from ipywidgets import Dropdown
import ipywidgets as widgets
from pandas.core.frame import DataFrame


def select_source_dir(path_dict, examples_dir=None):
    # Function for the button to select user input
    # Add apply selection to pass by ref 'path_dict'
    def get_user_selection(a):  # A default arg is needed here, I am guessing to pass self
        # Displays the current value of source_path
        path_dict['source_dir'] = dropbox.value
        # clear_output()
        display(path_dict['source_dir'])
        set_file_paths(dropbox.value)
        set_file_paths(dropbox.value)
        display(path_dict)

        # Execulte jump to next cell
        display(Javascript("var i = Jupyter.notebook.get_selected_index(); Jupyter.notebook.select(i+1);"))

    def is_valid_example(dir_value):
        file_list = ['location', 'account']
        for f in file_list:
            path = os.path.join(dir_value, f'{f}.csv')
            if not os.path.isfile(path):
                return False
        return True

    # Function to search and load files with standard names
    def set_file_paths(dir_value):
        file_list = ['location', 'account', 'ri_info', 'ri_scope']
        for f in file_list:
            path = os.path.join(dir_value, f'{f}.csv')
            path_dict[f'{f}_path'] = path if os.path.isfile(path) else None

    # Load default value and build examples list
    e = os.path.abspath(examples_dir)
    option_list = [os.path.join(e, o) for o in os.listdir(e) if os.path.isdir(os.path.join(e, o))]
    option_list.sort()

    # if 'examples_dir' is an example dir load it by default
    if is_valid_example(e):
        path_dict['source_dir'] = e
    else:
        path_dict['source_dir'] = os.path.join(e, 'uploaded')
    set_file_paths(path_dict['source_dir'])

    # creation of a widget dropdown object for directory selection
    dropbox = widgets.Dropdown(
        options=option_list,  # Object to iterate over
        description='Select:',  # User defined
        rows=10,  # The number of rows to display when showing the box
        interactive=True,  # This makes the box interactive,
    )

    # Button to click
    select_button = widgets.Button(
        description='Set source dir',  # User defined
        disabled=False
    )

    # Event Handlers
    select_button.on_click(get_user_selection)

    # display the UI
    ui = widgets.HBox([dropbox, select_button])  # pass an array of widgets to the ui
    display(ui)
    display("Default value: {}".format(path_dict['source_dir']))
    display(path_dict)



__all__ = [
    'PlatformBase',
    'PlatformList',
    'PlatformGet',
    'PlatformRun',
    'PlatformDelete',
]


import getpass
import io
import os
import json

from tabulate import tabulate
from mimetypes import guess_extension

from requests.exceptions import HTTPError

from ...platform_api.client import APIClient
from ...utils.exceptions import OasisException
from ...utils.defaults import API_EXAMPLE_AUTH

from ..base import ComputationStep


class PlatformBase(ComputationStep):
    """
    Base platform class to handle opening a client connection
    """
    step_params = [
        {'name': 'server_login_json', 'required': False, 'default': None, 'is_path': True,
            'pre_exist': False, 'help': 'Server login credentials json string'},
        {'name': 'server_url', 'default': 'http://localhost:8000', 'help': 'URL to Oasis Platform server, default is localhost'},
        {'name': 'server_version', 'default': 'v2', 'help': "Version prefix for OasisPlatform server, 'v1' = single server run, 'v2' = distributed on cluster"},
    ]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.server = self.open_connection()

    def load_credentials(self, login_arg):
        """
        Load credentials from JSON file or Prompt for username / password

        Options:
            1.'--server-login ./APIcredentials.json'
            2. Load credentials from default config file '-C oasislmf.json'
        """
        if isinstance(login_arg, str):
            with io.open(login_arg, encoding='utf-8') as f:
                return json.load(f)

        while True:
            user_response = input("Use simple JWT [Y/n]: ").lower().strip()
            if user_response in ["y", "yes"]:
                use_simple = True
                break
            elif user_response in ["n", "no"]:
                use_simple = False
                break

        self.logger.info('API Login:')
        api_login = {}
        if use_simple:
            api_login['username'] = input('Username: ')
            api_login['password'] = getpass.getpass('Password: ')
        else:
            api_login['client_id'] = input('Client ID: ')
            api_login['client_secret'] = getpass.getpass('Client Secret: ')

        return api_login

    def try_connection(self, fail_safe=True, **kwargs):
        """Helper to safely try connecting and return None if unauthorized."""
        try:
            return APIClient(
                api_url=self.server_url,
                api_ver=self.server_version,
                **kwargs
            )
        except OasisException as e:
            if not fail_safe:
                raise e
            orig_excep = e.original_exception
            if (isinstance(orig_excep, HTTPError) and orig_excep.response.status_code == 401):
                self.logger.debug("Login attempt failed, reason: Unauthorized (401) for credentials: %s", list(kwargs.keys()))
                return None
            elif isinstance(orig_excep, HTTPError) and orig_excep.response.status_code == 400 and "flow is disabled" in orig_excep.response.text.lower():
                self.logger.debug(
                    "Login attempt failed, reason: Validation Error (400) for credentials, invalid flow used, must be one of username/password or client_id/client_secret: %s", list(
                        kwargs.keys())
                )
                return None
            else:
                raise e  # Some other error – propagate

    def open_connection(self):
        """
        Attempts connection in this order:
        1. API_EXAMPLE_AUTH username/password
        2. API_EXAMPLE_AUTH client_id/client_secret
        3. Prompt or load credentials
        """
        if not isinstance(self.server_login_json, str):
            # 1. Try example username/password
            if 'username' in API_EXAMPLE_AUTH and 'password' in API_EXAMPLE_AUTH:
                conn = self.try_connection(
                    auth_type="simple",
                    username=API_EXAMPLE_AUTH['username'],
                    password=API_EXAMPLE_AUTH['password']
                )
                if conn:
                    return conn

            # 2. Try example client_id/client_secret
            if 'client_id' in API_EXAMPLE_AUTH and 'client_secret' in API_EXAMPLE_AUTH:
                conn = self.try_connection(
                    auth_type="oidc",
                    client_id=API_EXAMPLE_AUTH['client_id'],
                    client_secret=API_EXAMPLE_AUTH['client_secret']
                )
                if conn:
                    return conn

        # 3. Load credentials (file or prompt)
        self.logger.info("-- Authentication Required --")
        credentials = self.load_credentials(self.server_login_json)
        self.logger.info(f'Connecting to - {self.server_url}')
        if 'username' in credentials and 'password' in credentials:
            return self.try_connection(
                fail_safe=False,
                auth_type="simple",
                username=credentials['username'],
                password=credentials['password']
            )
        if 'client_id' in credentials and 'client_secret' in credentials:
            return self.try_connection(
                fail_safe=False,
                auth_type="oidc",
                client_id=credentials['client_id'],
                client_secret=credentials['client_secret']
            )
        raise OasisException(
            f"Error: No valid credentials provided for platform, current credential keys [{credentials.keys()}], must be one of username/password or client_id/client_secret"
        )

    def tabulate_json(self, json_data, items):
        table_data = dict()

        for i in items:
            table_data[i] = list()

        for m in json_data:
            for k in table_data:
                # will have link+data if dict returned
                if isinstance(m[k], dict):
                    table_data[k].append('Yes')

                # If none then no data
                elif m[k] is None:
                    table_data[k].append('-')

                # If URL then something linked to field
                elif isinstance(m[k], str):
                    if any(v in m[k] for v in ['http://', 'https://']):
                        table_data[k].append('Linked')
                    else:
                        table_data[k].append(m[k])

                # Fallback - add value as string
                else:
                    table_data[k].append(str(m[k]))
        return table_data

    def print_endpoint(self, attr, items):
        endpoint_obj = getattr(self.server, attr)
        self.logger.info(f'\nAvailable {attr}:')
        data = self.tabulate_json(endpoint_obj.get().json(), items)
        self.logger.info(tabulate(data, headers=items, tablefmt='psql'))
        return data

    def select_id(self, msg, valid_ids):
        while True:
            try:
                value = str(input(f'Select {msg} ID: '))
            except ValueError:
                self.logger.info('Invalid Response: {}'.format(value))
                continue
            except KeyboardInterrupt:
                return -1

            if value not in valid_ids:
                self.logger.info(f'id {value} not among the valid ids: {valid_ids} - ctrl-c to exit')
                continue
            else:
                break
        return int(value)


class PlatformList(PlatformBase):
    """ Return status and details from an Oasis Platform API server
    """
    step_params = PlatformBase.step_params + [
        {'name': 'models', 'flag': '-m', 'type': int, 'nargs': '+', 'help': 'List of model ids to print in detail'},
        {'name': 'portfolios', 'flag': '-p', 'type': int, 'nargs': '+', 'help': 'List of portfolio ids to print in detail'},
        {'name': 'analyses', 'flag': '-a', 'type': int, 'nargs': '+', 'help': 'List of analyses ids to print in detail'},
    ]

    def run(self):
        # Default to printing summary of API status
        if not any([self.models, self.portfolios, self.analyses]):
            self.print_endpoint('models', ['id', 'supplier_id', 'model_id', 'version_id'])
            self.print_endpoint('portfolios', ['id', 'name', 'location_file', 'accounts_file', 'reinsurance_info_file', 'reinsurance_scope_file'])
            self.print_endpoint('analyses', ['id', 'name', 'model', 'portfolio', 'status', 'input_file', 'output_file', 'run_log_file'])

        if self.models:
            for Id in self.models:
                msg = f'Model (id={Id}): \n'
                try:
                    rsp = self.server.models.get(Id)
                    self.logger.info(msg + json.dumps(rsp.json(), indent=4, sort_keys=True))
                except HTTPError as e:
                    self.logger.info(msg + e.response.text)

        if self.portfolios:
            for Id in self.portfolios:
                msg = f'Portfolio (id={Id}): \n'
                try:
                    rsp = self.server.portfolios.get(Id)
                    self.logger.info(msg + json.dumps(rsp.json(), indent=4, sort_keys=True))
                except HTTPError as e:
                    self.logger.info(msg + e.response.text)

        if self.analyses:
            for Id in self.analyses:
                msg = f'Analysis (id={Id}): \n'
                try:
                    rsp = self.server.analyses.get(Id)
                    self.logger.info(msg + json.dumps(rsp.json(), indent=4, sort_keys=True))
                except HTTPError as e:
                    self.logger.info(msg + e.response.text)


class PlatformRunInputs(PlatformBase):
    """ run generate inputs via the Oasis Platoform API
    """
    step_params = PlatformBase.step_params + [
        {'name': 'model_id', 'type': int, 'help': 'API `id` of a model to run an analysis with'},
        {'name': 'portfolio_id', 'type': int, 'help': 'API `id` of a portfolio to run an analysis with'},
        {'name': 'analysis_id', 'type': int, 'help': 'API `id` of an analysis to run'},

        {'name': 'analysis_settings_json', 'flag': '-a', 'is_path': True, 'pre_exist': True, 'help': 'Analysis settings JSON file path'},
        {'name': 'oed_location_csv', 'flag': '-x', 'is_path': True, 'pre_exist': True, 'help': 'Source location CSV file path'},
        {'name': 'oed_accounts_csv', 'flag': '-y', 'is_path': True, 'pre_exist': True, 'help': 'Source accounts CSV file path'},
        {'name': 'oed_info_csv', 'flag': '-i', 'is_path': True, 'pre_exist': True, 'help': 'Reinsurance info. CSV file path'},
        {'name': 'oed_scope_csv', 'flag': '-s', 'is_path': True, 'pre_exist': True, 'help': 'Reinsurance scope CSV file path'},
        {'name': 'currency_conversion_json', 'is_path': True, 'pre_exist': True, 'help': 'settings to perform currency conversion of oed files'},
        {'name': 'reporting_currency', 'type': str, 'help': 'currency to use in the results reported'},
    ]

    def run(self):
        # Run Input geneneration from ID
        if self.analysis_id:
            try:
                status = self.server.analyses.status(self.analysis_id)
                if status in ['RUN_QUEUED', 'RUN_STARTED']:
                    self.server.cancel_analysis(self.analysis_id)
                elif status in ['INPUTS_GENERATION_QUEUED', 'INPUTS_GENERATION_STARTED']:
                    self.server.cancel_generate(self.analysis_id)
                self.server.run_generate(self.analysis_id)
                return self.analysis_id
            except HTTPError as e:
                raise OasisException(f'Error running analysis ({self.analysis_id}) - {e}')

        # Create Portfolio and Ananlysis, then run
        if not (self.portfolio_id or self.oed_location_csv or self.oed_accounts_csv):
            raise OasisException('Error: At least one of the following inputs is required [portfolio_id, oed_location_csv, oed_accounts_csv]')

        # when no model is selected prompt user for choice
        if not self.model_id:
            models = self.server.models.get().json()
            model_count = len(models)

            if model_count < 1:
                raise OasisException(f'No models found in API: {self.server_url}')
            if model_count == 1:
                self.model_id = models[0]['id']
            if model_count > 1:
                models_table = self.print_endpoint('models', ['id', 'supplier_id', 'model_id', 'version_id'])
                self.model_id = self.select_id('models', models_table['id'])
            if self.model_id < 0:
                raise OasisException(' Model selection cancelled')

        # Select or create a portfilo
        if self.portfolio_id:
            portfolios = self.server.portfolios.get().json()
            if self.portfolio_id not in [p['id'] for p in portfolios]:
                raise OasisException(f'Portfolio "{self.portfolio_id}" not found in API: {self.server_url}')
        else:
            portfolio = self.server.upload_inputs(
                portfolio_id=None,
                location_fp=self.oed_location_csv,
                accounts_fp=self.oed_accounts_csv,
                ri_info_fp=self.oed_info_csv,
                ri_scope_fp=self.oed_scope_csv,
                currency_conversion_fp=self.currency_conversion_json,
                reporting_currency=self.reporting_currency
            )
            self.portfolio_id = portfolio['id']
        analysis = self.server.create_analysis(
            portfolio_id=self.portfolio_id,
            model_id=self.model_id,
            analysis_settings_fp=self.analysis_settings_json,
        )
        self.analysis_id = analysis['id']

        # Execure run
        self.server.run_generate(self.analysis_id)
        return self.analysis_id


class PlatformRunLosses(PlatformBase):
    """ run generate losses via the Oasis Platoform API
    """
    step_params = PlatformBase.step_params + [
        {'name': 'analysis_id', 'type': int, 'required': True, 'help': 'API `id` of an analysis to run'},
        {'name': 'output_dir', 'flag': '-o', 'is_path': True, 'pre_exist': True,
            'help': 'Output data directory for results data (absolute or relative file path)', 'default': './'},
        {'name': 'analysis_settings_json', 'flag': '-a', 'is_path': True, 'pre_exist': True, 'help': 'Analysis settings JSON file path'},
    ]

    def run(self):
        self.server.run_analysis(self.analysis_id, self.analysis_settings_json)
        self.server.download_output(self.analysis_id, self.output_dir)


class PlatformRun(PlatformBase):
    """ End to End - run model via the Oasis Platoform API
    """
    chained_commands = [PlatformRunInputs, PlatformRunLosses]

    def run(self):
        self.kwargs['analysis_id'] = PlatformRunInputs(**self.kwargs).run()
        PlatformRunLosses(**self.kwargs).run()


class PlatformDelete(PlatformBase):
    """ Delete either a 'model', 'portfolio' or an 'analysis' from the API's Database
    """
    step_params = PlatformBase.step_params + [
        {'name': 'models', 'flag': '-m', 'type': int, 'nargs': '+', 'help': 'List of model ids to Delete.'},
        {'name': 'portfolios', 'flag': '-p', 'type': int, 'nargs': '+', 'help': 'List of Portfolio ids to Delete'},
        {'name': 'analyses', 'flag': '-a', 'type': int, 'nargs': '+', 'help': 'List of Analyses ids to Detele'},
    ]

    def delete_list(self, attr, id_list):
        if not all(isinstance(ID, int) for ID in id_list):
            raise OasisException(f"Invalid input, '{attr}', must be a list of type Int, not {id_list}")

        api_endpoint = getattr(self.server, attr)
        for Id in id_list:
            try:
                if api_endpoint.delete(Id).ok:
                    self.logger.info(f'Deleted {attr}_id={Id}')
            except HTTPError as e:
                self.logger.error('Delete error {}_id={} - {}'.format(attr, Id, e))
                continue

    def run(self):

        if not any([self.models, self.portfolios, self.analyses]):
            raise OasisException("""Select item(s) to delete, list of either:
                --models MODEL_ID [MODEL_ID ... n],
                --portfolios PORTFOLIOS_ID [PORTFOLIO_ID ... n]
                --analyses ANALYSES_ID [ANALYSES_ID ... n]
                """)

        if self.models:
            self.delete_list('models', self.models)
        if self.portfolios:
            self.delete_list('portfolios', self.portfolios)
        if self.analyses:
            self.delete_list('analyses', self.analyses)


class PlatformGet(PlatformBase):
    """ Download file(s) from the api
    """
    step_params = PlatformBase.step_params + [
        {'name': 'output_dir', 'flag': '-o', 'is_path': True, 'pre_exist': True,
            'help': 'Output data directory for results data (absolute or relative file path)', 'default': './'},
        # Files for models object
        {'name': 'model_settings', 'type': int, 'nargs': '+', 'help': 'Model ids to download settings file.'},
        {'name': 'model_versions', 'type': int, 'nargs': '+', 'help': 'Model ids to download versions file'},
        # Files from portfolio
        {'name': 'portfolio_location_file', 'type': int, 'nargs': '+', 'help': 'Portfolio ids to download Location file'},
        {'name': 'portfolio_accounts_file', 'type': int, 'nargs': '+', 'help': 'Portfolio ids to download Accounts file'},
        {'name': 'portfolio_reinsurance_scope_file', 'type': int, 'nargs': '+', 'help': 'Portfolio ids to download RI scope file.'},
        {'name': 'portfolio_reinsurance_info_file', 'type': int, 'nargs': '+', 'help': 'Portfolio ids to download RI info file.'},
        # Files from an analyses
        {'name': 'analyses_settings_file', 'type': int, 'nargs': '+', 'help': 'Analyses ids to download settings file'},
        {'name': 'analyses_run_traceback_file', 'type': int, 'nargs': '+', 'help': 'Analyses ids to download traceback logs'},
        {'name': 'analyses_run_log_file', 'type': int, 'nargs': '+', 'help': 'Analyses ids to download Ktools run logs'},
        {'name': 'analyses_input_generation_traceback_file', 'type': int, 'nargs': '+',
            'help': 'Analyses ids to download input_generation traceback logs'},
        {'name': 'analyses_input_file', 'type': int, 'nargs': '+', 'help': 'Analyses ids to download Generated inputs tar'},
        {'name': 'analyses_output_file', 'type': int, 'nargs': '+', 'help': 'Analyses ids to download Output losses tar'},
        {'name': 'analyses_summary_levels_file', 'type': int, 'nargs': '+', 'help': 'Analyses ids to download summary levels file'},
        {'name': 'analyses_lookup_validation_file', 'type': int, 'nargs': '+', 'help': 'Analyses ids to download exposure summary'},
        {'name': 'analyses_lookup_success_file', 'type': int, 'nargs': '+', 'help': 'Analyses ids to download successful lookups'},
        {'name': 'analyses_lookup_errors_file', 'type': int, 'nargs': '+', 'help': 'Analyses ids to download summary of failed lookups'},
    ]

    def extract_args(self, param_suffix):
        return {k.replace(param_suffix, ''): v for k, v in self.kwargs.items() if v and param_suffix in k}

    def download(self, collection, req_files, chuck_size=1024):
        collection_obj = getattr(self.server, collection)
        for File in req_files:
            resource = getattr(collection_obj, File)
            for ID in req_files[File]:
                try:
                    r = resource.get(ID)

                    ext = guess_extension(r.headers['content-type'].partition(';')[0].strip())
                    filename = os.path.join(self.output_dir, f'{ID}_{collection}_{File}{ext}')

                    with io.open(filename, 'wb') as f:
                        for chunk in r.iter_content(chunk_size=chuck_size):
                            f.write(chunk)
                    self.logger.info(f'Downloaded: {File} from {collection}_id={ID} "{filename}"')
                except HTTPError as e:
                    self.logger.error('Download failed: - {}'.format(e))

    def run(self):
        model_files = self.extract_args('model_')
        portfolio_files = self.extract_args('portfolio_')
        analyses_files = self.extract_args('analyses_')

        # Check that at least one option is given
        if not any([model_files, portfolio_files, analyses_files]):
            raise OasisException('Select file for download e.g. "--analyses_output <id_1> .. <id_n>"')

        if model_files:
            self.download('models', model_files)
        if portfolio_files:
            self.download('portfolios', portfolio_files)
        if analyses_files:
            self.download('analyses', analyses_files)



from oasislmf.pytools.elt.data import MELT_dtype, MELT_fmt, MELT_headers, QELT_dtype, QELT_fmt, QELT_headers, SELT_dtype, SELT_fmt, SELT_headers
from oasislmf.pytools.plt.data import MPLT_dtype, MPLT_fmt, MPLT_headers, QPLT_dtype, QPLT_fmt, QPLT_headers, SPLT_dtype, SPLT_fmt, SPLT_headers


VALID_EXT = ["csv", "parquet", "bin", None]

KAT_SELT = 0
KAT_MELT = 1
KAT_QELT = 2
KAT_SPLT = 3
KAT_MPLT = 4
KAT_QPLT = 5

KAT_MAP = {
    KAT_SELT: {
        "name": "SELT",
        "headers": SELT_headers,
        "dtype": SELT_dtype,
        "fmt": SELT_fmt,
    },
    KAT_MELT: {
        "name": "MELT",
        "headers": MELT_headers,
        "dtype": MELT_dtype,
        "fmt": MELT_fmt,
    },
    KAT_QELT: {
        "name": "QELT",
        "headers": QELT_headers,
        "dtype": QELT_dtype,
        "fmt": QELT_fmt,
    },
    KAT_SPLT: {
        "name": "SPLT",
        "headers": SPLT_headers,
        "dtype": SPLT_dtype,
        "fmt": SPLT_fmt,
    },
    KAT_MPLT: {
        "name": "MPLT",
        "headers": MPLT_headers,
        "dtype": MPLT_dtype,
        "fmt": MPLT_fmt,
    },
    KAT_QPLT: {
        "name": "QPLT",
        "headers": QPLT_headers,
        "dtype": QPLT_dtype,
        "fmt": QPLT_fmt,
    },
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Oasis LMF documentation build configuration file, created by
# sphinx-quickstart on Thu Feb  1 12:22:30 2018.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

import datetime
# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
import os
import subprocess
import sys

from sphinx.ext import autodoc

ROOT_PATH = os.path.join(os.path.dirname(__file__), os.path.pardir, os.path.pardir)

sys.path.append(ROOT_PATH)
from oasislmf import __version__  # noqa
from oasislmf.cli.model import RunCmd
from oasislmf.computation.run.model import RunModel
from oasislmf.computation.run.exposure import RunExposure
from oasislmf.computation.run.platform import PlatformBase, PlatformRun

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'autoapi.extension',
]


# API Generation
autoapi_dirs = ['../../oasislmf']
autoapi_options = [
    "members",
    # "inherited-members",  # errors
    "undoc-members",
    "show-inheritance",
    "show-module-summary",
]
autoapi_keep_files = False

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'Oasis LMF'
author = u'Oasis LMF'
year_now = datetime.date.today().year
copyright = str(year_now) + ' Oasis LMF'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = __version__
# The full version, including alpha/beta/rc tags.
release = __version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'friendly'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'furo'
html_title = "Oasis LMF Documentation"

html_css_files = [
    'https://fonts.googleapis.com/css?family=Raleway',
]
# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/OasisLMF",
            "html": """
                <svg stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 16 16">
                    <path fill-rule="evenodd" d="M8 0c4.42 0 8 3.58 8 8a8.013 8.013 0 0 1-5.45 7.59c-.4.08-.55-.17-.55-.38 0-.27.01-1.13.01-2.2 0-.75-.25-1.23-.54-1.48 1.78-.2 3.65-.88 3.65-3.95 0-.88-.31-1.59-.82-2.15.08-.2.36-1.02-.08-2.12 0 0-.67-.22-2.2.82-.64-.18-1.32-.27-2-.27-.68 0-1.36.09-2 .27-1.53-1.03-2.2-.82-2.2-.82-.44 1.1-.16 1.92-.08 2.12-.51.56-.82 1.28-.82 2.15 0 3.06 1.86 3.75 3.64 3.95-.23.2-.44.55-.51 1.07-.46.21-1.61.55-2.33-.66-.15-.24-.6-.83-1.23-.82-.67.01-.27.38.01.53.34.19.73.9.82 1.13.16.45.68 1.31 2.69.94 0 .67.01 1.3.01 1.49 0 .21-.15.45-.55.38A7.995 7.995 0 0 1 0 8c0-4.42 3.58-8 8-8Z"></path>
                </svg>
            """,
            "class": "",
        },
    ],

    "light_css_variables": {
        "color-brand-primary": " #862633",
        "color-brand-content": "#d22630",
        "font-stack": "Raleway, sans-serif",
        "font-stack--monospace": "Courier, monospace",
    }
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
# html_sidebars = {
#     '**': [
#         'about.html',
#         'navigation.html',
#         'relations.html',  # needs 'show_related': True theme option to display
#         'searchbox.html',
#         'donate.html',
#         'versioning.html'
#     ]
# }


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'OasisLMFdoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'OasisLMF.tex', 'Oasis LMF Documentation',
     'Oasis Loss Modelling Framework', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'oasislmf', 'Oasis LMF Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'OasisLMF', 'Oasis LMF Documentation',
     author, 'OasisLMF', 'One line description of project.',
     'Miscellaneous'),
]


# -- Options for Epub output ----------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project
epub_author = author
epub_publisher = author
epub_copyright = copyright

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'python': ('https://docs.python.org/', None)}


def list_options():
    # Get params from RunCmd
    cmd_opts = RunCmd().arg_parser._actions
    cmd_opts.pop(0)

    # Retrieve params from all other relevant subcommands
    runmodel_info = RunModel.get_params()
    runexposure_info = RunExposure.get_params()
    platformbase_info = PlatformBase.get_params()
    platformrun_info = PlatformRun.get_params()

    # Populate a dictionary with all the params from subcommands
    all_params = {param['name']: param for param in runmodel_info}
    all_params.update({param['name']: param for param in runexposure_info})
    all_params.update({param['name']: param for param in platformbase_info})
    all_params.update({param['name']: param for param in platformrun_info})

    rst_output = []

    # Create a dictionary from cmd_opts
    cmd_opts_dict = {}
    for opt in cmd_opts:
        param_name = opt.dest
        expected_type = opt.type if opt.type else 'string'
        cmd_opts_dict[param_name] = {
            'help': opt.help,
            'default': opt.default,
            'expected_type': expected_type,
            'option_strings': opt.option_strings
        }

    # Update cmd_opts with details from subcommand params
    for param_name, param_entry in all_params.items():
        if param_name in cmd_opts_dict:
            cmd_opt = cmd_opts_dict[param_name]
            if cmd_opt['default'] is None and 'default' in param_entry:
                cmd_opt['default'] = param_entry['default']
            if 'type' in param_entry:
                cmd_opt['expected_type'] = param_entry['type']
            if param_entry.get('is_path', False):
                cmd_opt['expected_type'] = "path"
        else:
            cmd_opts_dict[param_name] = {
                'help': param_entry.get('help', ''),
                'default': param_entry.get('default', None),
                'expected_type': param_entry.get('type', 'string'),
                'option_strings': []
            }
            if param_entry.get('is_path', False):
                cmd_opts_dict[param_name]['expected_type'] = "path"

    # Remove options that are exclusively passed as data during execution
    cmd_opts_dict = {k: v for k, v in cmd_opts_dict.items() if k not in ['profile_loc', 'profile_acc', 'profile_fm_agg']}

    # Reformat expected_type
    for cmd_opt in cmd_opts_dict.values():
        expected_type_str = str(cmd_opt['expected_type'])
        if "<function str2bool at" in expected_type_str:
            cmd_opt['expected_type'] = "boolean (yes/no, true/false t/f, y/n, or 1/0)"
        elif "<oasislmf.utils.path.PathCleaner" in expected_type_str:
            cmd_opt['expected_type'] = "path"
        elif "<class 'int'>" in expected_type_str:
            cmd_opt['expected_type'] = "integer"

    # Generate RST output
    rst_output = []
    for param_name, cmd_opt in cmd_opts_dict.items():
        rst_output.extend([
            f'{param_name}',
            '=' * len(param_name),
            '',
            f'Description: {cmd_opt["help"]}',
            '',
            f'Expected type: {cmd_opt["expected_type"]}',
            '',
            f'Default value: ``{cmd_opt["default"]}``',
            '',
        ])

    with open('./source/generated_options.rst', 'w') as f:
        f.write('\n'.join(rst_output))


class CliDocumenter(autodoc.ClassDocumenter):
    objtype = "cli"

    # do not indent the content
    content_indent = ""

    # do not add a header to the docstring
    def add_directive_header(self, sig):
        pass


def setup(app):
    app.add_autodocumenter(CliDocumenter)
    list_options()



#!/usr/bin/env python3

"""
Installs a version of the MDK package from a given branch in the OasisLMF
GitHub repository, and does an end-to-end runtime test of an OasisLMF-managed
model repository on GitHub using a small sample dataset. Runtime options
include 'gul' for ground up loss (GUL) only, "fm" for insured loss, or 'ri' for
reinsurance losses. In practice, the test model repository will generally be
PiWind.
"""

import argparse
import copy
import io
import json
import os
import shutil
import subprocess
import sys

from subprocess import (
    CalledProcessError,
    run,
)

from oasislmf.utils.path import setcwd


class MdkModelRunException(Exception):
    pass


def get_default_pip_path():
    default_pip_path = ''

    cmd_str = 'which pip'
    try:
        resp = run(cmd_str.split(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT, check=True).stdout
    except CalledProcessError:
        return
    else:
        if not resp:
            raise MdkModelRunException('No default pip path found!')
        default_pip_path = resp.decode('utf-8').strip()

    return default_pip_path


def parse_args():
    parser = argparse.ArgumentParser(
        description='Installs a version of the MDK package from a given branch in the OasisLMF '
                    'GitHub repository, and does an end-to-end runtime test of an OasisLMF-managed '
                    'model repository on GitHub using a small sample dataset. Runtime options '
                    'include "gul" for ground up loss (GUL) only, "fm" for insured loss, or "ri" for '
                    'reinsurance losses. In practice, the test model repository will generally be'
                    'PiWind'
    )

    parser.add_argument('-b', '--mdk-repo-branch', default='develop',
                        help='Target branch in the MDK package GitHub repository to build the package from')

    parser.add_argument('-m', '--model-repo-name', default='OasisPiWind',
                        help='Target model GitHub repository name (must be an OasisLMF managed repository)')

    parser.add_argument('-r', '--model-repo-branch', default='master', help='Target branch in the model GitHub repository to clone')

    parser.add_argument('-t', '--clone-target', default=os.path.abspath('.'),
                        help='Local parent folder in which to clone the model repository - default is script run directory')

    parser.add_argument('-g', '--git-transfer-protocol', default='ssh', help='Git transfer protocol - https" or "ssh"')

    parser.add_argument('-p', '--pip-path', default=get_default_pip_path(), help='pip path')

    parser.add_argument('-d', '--model-run-mode', default='ri', help='Model run mode - `gul` for GUL only, `fm` for GUL + FM, `ri` for GUL + FM + RI')

    parser.add_argument('-n', '--no-cleanup', action='store_true', default=False,
                        help='Whether to cleanup installed MDK installed package and model repository')

    args = vars(parser.parse_args())

    if not os.path.isabs(args['clone_target']):
        args['clone_target'] = os.path.abspath(args['clone_target'])

    args['model_run_mode'] = args['model_run_mode'].lower()
    if args['model_run_mode'] not in ['gul', 'fm', 'ri']:
        args['model_run_mode'] = 'ri'

    return args


def pkg_exists(pkg_name):
    cmd_str = 'pip freeze | grep {}'.format(pkg_name)
    try:
        resp = run(cmd_str.split(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT, check=True).stdout
    except CalledProcessError:
        return False
    else:
        if not resp:
            raise MdkModelRunException('Error while trying to determine whether {} is installed'.format(pkg_name))
        return pkg_name in resp.decode('utf-8').strip()


def pip_uninstall(pkg_name, options_str='-y -v', pip_path=get_default_pip_path()):
    if pkg_exists(pkg_name):
        cmd_str = 'pip uninstall {} {}'.format(options_str, pkg_name)
        run(cmd_str.split(), check=True)


def pip_install(pkg_name_or_branch_uri, options_str='-v', pip_path=get_default_pip_path()):
    pkg_name = pkg_name_or_branch_uri.split('=')[-1]

    if pkg_exists(pkg_name):
        pip_uninstall(pkg_name)

    cmd_str = '{} install {} {}'.format(pip_path, options_str, pkg_name_or_branch_uri)
    run(cmd_str.split(), check=True)


def clone_repo(repo_name, target, repo_branch='master', user_or_org_name='OasisLMF', home=os.getcwd(), transfer_protocol='ssh'):
    if not os.path.exists(target):
        os.mkdir(target)

    repo_target = os.path.join(target, repo_name)
    if os.path.exists(repo_target):
        shutil.rmtree(repo_target)

    with setcwd(target):
        repo_url = (
            'git+ssh://git@github.com/{}/{}'.format(user_or_org_name, repo_name) if transfer_protocol == 'ssh'
            else 'https://github.com/{}/{}'.format(user_or_org_name, repo_name)
        )

        options_str = '-b {} --single-branch'.format(repo_branch)

        cmd_str = 'git clone {} {}'.format(options_str, repo_url)

        run(cmd_str.split(), check=True)


def apply_model_run_mode(model_run_mode, model_mdk_config_fp, as_dict=False):
    with io.open(model_mdk_config_fp, 'r', encoding='utf-8') as f:
        model_mdk_config = json.load(f)

    _model_mdk_config = copy.deepcopy(model_mdk_config)

    if model_run_mode == 'gul':
        for k in model_mdk_config:
            if 'accounts' in k or 'fm' in k or 'ri' in k:
                _model_mdk_config.pop(k)
    elif model_run_mode == 'fm':
        for k in model_mdk_config:
            if 'ri' in k:
                _model_mdk_config.pop(k)

    if as_dict:
        return _model_mdk_config

    with io.open(model_mdk_config_fp, 'w', encoding='utf-8') as f:
        json.dump(_model_mdk_config, f, indent=4, sort_keys=True)


def run_model(model_mdk_config_fp, model_run_dir=os.path.abspath('.')):
    cmd_str = 'oasislmf model run -C {} -r {}'.format(model_mdk_config_fp, model_run_dir)
    run(cmd_str.split(), check=True)


def print_model_dir_tree(model_run_dir, options_str='-h'):
    cmd_str = 'tree {} {}'.format(options_str, model_run_dir)
    run(cmd_str.split(), check=True)


def model_run_ok(model_run_dir, model_run_mode):

    def _is_non_empty_file(fp, substr_match=False, is_dir=False):
        if not substr_match:
            return (os.path.isfile(fp) if not is_dir else os.path.isdir(fp)) and os.path.getsize(fp) > 0
        else:
            substr, dir_name, dir_contents = os.path.basename(fp), os.path.dirname(fp), os.listdir(os.path.dirname(fp))
            try:
                fn = [fn for fn in dir_contents if substr.lower() in fn.lower()][0]
            except (AttributeError, IndexError):
                return False
            _fp = os.path.join(dir_name, fn)
            return os.path.getsize(_fp) > 0

    ri = model_run_mode == 'ri'

    assert (_is_non_empty_file(model_run_dir, is_dir=True))

    assert (_is_non_empty_file(os.path.join(model_run_dir, 'analysis_settings.json')))
    assert (_is_non_empty_file(os.path.join(model_run_dir, 'input'), is_dir=True))
    assert (_is_non_empty_file(os.path.join(model_run_dir, 'output'), is_dir=True))
    assert (_is_non_empty_file(os.path.join(model_run_dir, 'static'), is_dir=True))
    assert (_is_non_empty_file(os.path.join(model_run_dir, 'work'), is_dir=True))
    assert (_is_non_empty_file(os.path.join(model_run_dir, 'run_kernel.sh')))

    direct_csv_inputs_fp = os.path.join(model_run_dir, 'input', 'csv') if not ri else os.path.join(model_run_dir, 'input')

    try:
        assert (_is_non_empty_file(os.path.join(direct_csv_inputs_fp, 'srcexp'), substr_match=True))
    except AssertionError:
        assert (_is_non_empty_file(os.path.join(direct_csv_inputs_fp, 'sourceloc'), substr_match=True))

    assert (_is_non_empty_file(os.path.join(direct_csv_inputs_fp, 'keys'), substr_match=True))
    assert (_is_non_empty_file(os.path.join(direct_csv_inputs_fp, 'keys-errors'), substr_match=True))

    assert (_is_non_empty_file(os.path.join(direct_csv_inputs_fp, 'items.csv')))
    assert (_is_non_empty_file(os.path.join(direct_csv_inputs_fp, 'coverages.csv')))
    assert (_is_non_empty_file(os.path.join(direct_csv_inputs_fp, 'gulsummaryxref.csv')))

    bin_inputs_fp = os.path.join(model_run_dir, 'input')
    assert (_is_non_empty_file(os.path.join(bin_inputs_fp, 'occurrence.bin')))
    assert (_is_non_empty_file(os.path.join(bin_inputs_fp, 'events.bin')))
    assert (_is_non_empty_file(os.path.join(bin_inputs_fp, 'items.bin')))
    assert (_is_non_empty_file(os.path.join(bin_inputs_fp, 'coverages.bin')))
    assert (_is_non_empty_file(os.path.join(bin_inputs_fp, 'gulsummaryxref.bin')))

    outputs_fp = os.path.join(model_run_dir, 'output')
    assert (_is_non_empty_file(os.path.join(outputs_fp, 'gul_S1_aalcalc.csv')))
    assert (_is_non_empty_file(os.path.join(outputs_fp, 'gul_S1_eltcalc.csv')))
    assert (_is_non_empty_file(os.path.join(outputs_fp, 'gul_S1_leccalc_full_uncertainty_aep.csv')))
    assert (_is_non_empty_file(os.path.join(outputs_fp, 'gul_S1_leccalc_full_uncertainty_oep.csv')))

    if model_run_mode in ['fm', 'ri']:
        try:
            assert (_is_non_empty_file(os.path.join(direct_csv_inputs_fp, 'srcacc'), substr_match=True))
        except AssertionError:
            assert (_is_non_empty_file(os.path.join(direct_csv_inputs_fp, 'sourceacc'), substr_match=True))

        assert (_is_non_empty_file(os.path.join(direct_csv_inputs_fp, 'fm_programme.csv')))
        assert (_is_non_empty_file(os.path.join(direct_csv_inputs_fp, 'fm_profile.csv')))
        assert (_is_non_empty_file(os.path.join(direct_csv_inputs_fp, 'fm_policytc.csv')))
        assert (_is_non_empty_file(os.path.join(direct_csv_inputs_fp, 'fm_xref.csv')))
        assert (_is_non_empty_file(os.path.join(direct_csv_inputs_fp, 'fmsummaryxref.csv')))

        assert (_is_non_empty_file(os.path.join(bin_inputs_fp, 'fm_programme.bin')))
        assert (_is_non_empty_file(os.path.join(bin_inputs_fp, 'fm_profile.bin')))
        assert (_is_non_empty_file(os.path.join(bin_inputs_fp, 'fm_policytc.bin')))
        assert (_is_non_empty_file(os.path.join(bin_inputs_fp, 'fm_xref.bin')))
        assert (_is_non_empty_file(os.path.join(bin_inputs_fp, 'fmsummaryxref.bin')))

        assert (_is_non_empty_file(os.path.join(outputs_fp, 'il_S1_aalcalc.csv')))
        assert (_is_non_empty_file(os.path.join(outputs_fp, 'il_S1_eltcalc.csv')))
        assert (_is_non_empty_file(os.path.join(outputs_fp, 'il_S1_leccalc_full_uncertainty_aep.csv')))
        assert (_is_non_empty_file(os.path.join(outputs_fp, 'il_S1_leccalc_full_uncertainty_oep.csv')))

        if model_run_mode == 'ri':
            assert (_is_non_empty_file(os.path.join(model_run_dir, 'ri_layers'), substr_match=True))
            assert (_is_non_empty_file(os.path.join(model_run_dir, 'RI'), substr_match=True))

    return True


def cleanup(pip_path=get_default_pip_path(), mdk_pkg_name='oasislmf', model_run_dir=None):
    pip_uninstall(mdk_pkg_name, options_str='-v -y', pip_path=pip_path)
    if model_run_dir:
        shutil.rmtree(model_run_dir)


if __name__ == "__main__":

    args = parse_args()

    print('\nProcessing script arguments: {}'.format(json.dumps(args, indent=4, sort_keys=True)))

    if not args['pip_path']:
        raise MdkModelRunException('pip path could not be determined and/or no pip path provided when calling the script')

    if args['git_transfer_protocol'] not in ['https', 'ssh']:
        args['git_transfer_protocol'] = 'ssh'

    pkg_uri = 'git+{}://git@github.com/OasisLMF/OasisLMF.git@{}#egg=oasislmf'.format(args['git_transfer_protocol'], args['mdk_repo_branch'])

    print('\nInstalling MDK package {}'.format(pkg_uri))

    try:
        pip_install(pkg_uri, pip_path=args['pip_path'])
    except CalledProcessError as e:
        raise MdkModelRunException('\nError trying to pip install package: {}'.format(e))

    print('\nMDK package successfully installed from branch {}'.format(args['mdk_repo_branch']))

    print('\nCloning model repository {} (branch {}) in {}\n'.format(args['model_repo_name'], args['model_repo_branch'], args['clone_target']))

    try:
        clone_repo(args['model_repo_name'], args['clone_target'], repo_branch=args['model_repo_branch'],
                   transfer_protocol=args['git_transfer_protocol'])
    except CalledProcessError as e:
        raise MdkModelRunException('\nError while trying to clone {} repository: {}\n'.format(args['model_repo_name'], e))

    print('\n{} successfully cloned in {}'.format(args['model_repo_name'], args['clone_target']))

    local_model_repo_fp = os.path.join(args['clone_target'], args['model_repo_name'])
    model_mdk_config_fp = os.path.join(local_model_repo_fp, 'oasislmf.json')

    print('\nAdjusting {} MDK config. file to suit model run mode "{}"'.format(args['model_repo_name'], args['model_run_mode'].upper()))
    apply_model_run_mode(args['model_run_mode'], model_mdk_config_fp)

    model_run_dir = os.path.join(local_model_repo_fp, 'test-run')
    print(
        '\nRunning {} end-to-end via MDK using config. file {} - model run dir. is {}\n'.format(args['model_repo_name'], model_mdk_config_fp, model_run_dir))

    try:
        run_model(model_mdk_config_fp, model_run_dir=model_run_dir)
    except CalledProcessError as e:
        raise MdkModelRunException('\nError while trying to run {} via MDK: {}'.format(args['model_repo_name'], e))

    print('\nModel run OK')

    print('\nModel run directory has the following structure\n')
    try:
        print_model_dir_tree(model_run_dir)
    except CalledProcessError:
        print('\nError getting or printing model run directory tree - skipping step')

    print('\nChecking correctess of model run directory')
    try:
        model_run_ok(model_run_dir, args['model_run_mode'])
    except AssertionError:
        print('\nModel run error - missing, incorrect or incomplete files in model run dir. {}'.format(model_run_dir))
        sys.exit(1)

    print('\nModel run directory has the expected structure')

    print('\nModel run completed successfully')

    if not args['no_cleanup']:
        print('\nCleaning up - removing MDK package install and test model repository {}'.format(args['model_repo_name']))
        try:
            cleanup(pip_path=args['pip_path'], model_run_dir=local_model_repo_fp)
        except CalledProcessError as e:
            print('\nError cleaning up: {}'.format(e))

    sys.exit(0)



#!/usr/bin/env python

import argparse

import pandas as pd

parser = argparse.ArgumentParser(
    usage='use "%(prog)s --help" for more information',
    formatter_class=argparse.RawTextHelpFormatter  # for multi-line help text
)

parser.add_argument('-p', '--parquet-path', help='path to parquet file', type=str, required=True)
parser.add_argument('-c', '--csv-path', help='path to csv file', type=str, required=True)
parser.add_argument('-s', '--sort', help='sort by string, int and categorical columns before check', action='store_true')
parser.add_argument('--check-exact', help='Whether to compare number exactly.', action='store_true')
parser.add_argument('--cols-case-sensitive', help='Lower cased column names before compare', action='store_true')
parser.add_argument('--raise-error', help='Raise AssertionError exception if DataFrames differ, default is to print the exception message', action='store_true')
parser.add_argument('--rtol', help='relative tolerance. Only used when check_exact is False', default=1e-05, action='store', type=float)
parser.add_argument('--atol', help='Absolute tolerance. Only used when check_exact is False', default=1e-08, action='store', type=float)


def is_sorting_type(dtype):
    return (pd.api.types.is_string_dtype(dtype) or
            pd.api.types.is_categorical_dtype(dtype) or
            pd.api.types.is_integer_dtype(dtype))


def compare(parquet_path, csv_path, sort, cols_case_sensitive, raise_error, **kwargs):
    parquet_df = pd.read_parquet(parquet_path)
    dtypes = parquet_df.dtypes.apply(lambda x: x.name).to_dict()
    csv_df = pd.read_csv(csv_path, dtype=dtypes)

    if not cols_case_sensitive:
        parquet_df.columns = parquet_df.columns.str.lower()
        csv_df.columns = csv_df.columns.str.lower()

    if sort:
        sort_column = [column for column, dtype in dtypes.items() if is_sorting_type(dtype)]
        parquet_df = parquet_df.sort_values(by=sort_column)
        csv_df = csv_df.sort_values(by=sort_column)

    try:
        pd.testing.assert_frame_equal(parquet_df, csv_df, **kwargs)
    except AssertionError as e:
        if raise_error:
            raise
        else:
            print(e)


if __name__ == '__main__':
    # example:
    # python ~/OasisLMF/validation/utils/compare_parquet_csv.py -p ~/OasisPiWind/tests/inputs/SourceLocOEDPiWind.parquet  -c ~/OasisPiWind/tests/inputs/SourceLocOEDPiWind.csv

    kwargs = vars(parser.parse_args())
    compare(**kwargs)



from requests import Session
from requests.adapters import HTTPAdapter
from requests_toolbelt import MultipartEncoder
from requests.exceptions import (
    ConnectionError,
    HTTPError,
    ReadTimeout,
)

from posixpath import join as urljoin
import time
import os
import logging

from ..utils.exceptions import OasisException


class APISession(Session):
    def __init__(
        self,
        api_url,
        auth_type=None,
        username=None,
        password=None,
        client_id=None,
        client_secret=None,
        access_token=None,
        refresh_token=None,
        timeout=25,
        retries=5,
        retry_delay=1,
        request_interval=0.02,
        logger=None,
        **kwargs
    ):
        super(APISession, self).__init__(**kwargs)
        self.logger = logger or logging.getLogger(__name__)

        # Extended class vars
        self.tkn_access = None
        self.tkn_refresh = None
        self.url_base = urljoin(api_url, '')
        self.timeout = timeout
        self.retry_max = 0
        self.retry_delay = retry_delay
        self.request_interval = request_interval
        self.headers = {
            'authorization': '',
            'accept': 'application/json',
            'content-type': 'application/json',
        }
        self.adapters.clear()
        self.mount(self.url_base, HTTPAdapter(max_retries=self.retry_max))

        # Check connectivity & authentication
        self.health_check()
        self.retry_max = retries
        self.auth_type = auth_type
        if auth_type == "simple" and username and password:
            self.auth_credentials = {"username": username, "password": password}
        elif auth_type == "oidc" and client_id and client_secret:
            self.auth_credentials = {"client_id": client_id, "client_secret": client_secret}
        elif auth_type == "token" and access_token and refresh_token:
            self.auth_credentials = {"access_token": access_token, "refresh_token": refresh_token}
        else:
            raise OasisException(
                f"Missing credentials for auth_type {auth_type}: must provide either username/password or client_id/client_secret or access_token/refresh_token.")
        self.__get_access_token()

    def __get_access_token(self):
        if self.auth_type == "token":
            self.tkn_access = self.auth_credentials["access_token"]
            self.tkn_refresh = self.auth_credentials["refresh_token"]
            return
        try:
            url = urljoin(self.url_base, 'access_token/')
            r = self.post(url, json=self.auth_credentials)
            r.raise_for_status()
            self.tkn_access = r.json()['access_token']
            if self.auth_type == "simple":
                self.tkn_refresh = r.json()['refresh_token']
            else:
                self.tkn_refresh = self.tkn_access
            self.headers['authorization'] = 'Bearer {}'.format(self.tkn_access)
            return
        except (TypeError, AttributeError, BytesWarning, HTTPError, ConnectionError, ReadTimeout) as e:
            err_msg = 'Authentication Error'
            raise OasisException(err_msg, e)

    def _refresh_token(self):
        if self.auth_type == "oidc":
            self.__get_access_token()
            return
        try:
            self.headers['authorization'] = 'Bearer {}'.format(self.tkn_refresh)
            url = urljoin(self.url_base, 'refresh_token/')
            r = super(APISession, self).post(url, timeout=self.timeout)
            r.raise_for_status()

            self.tkn_access = r.json()['access_token']
            if 'refresh_token' in r.json():
                self.logger.debug("Refreshing access token")
                self.tkn_refresh = r.json()['refresh_token']
            self.headers['authorization'] = 'Bearer {}'.format(self.tkn_access)
            return
        except (TypeError, AttributeError, BytesWarning, HTTPError, ConnectionError, ReadTimeout) as e:
            err_msg = 'Authentication Error'
            raise OasisException(err_msg, e)

    def unrecoverable_error(self, error, msg=None):
        err_r = error.response
        err_msg = 'api error: {}, url: {}, msg: {}'.format(err_r.status_code, err_r.url, err_r.text)
        if msg:
            self.logger.error(msg)
        raise OasisException(err_msg, error)

    # Connection Error Handlers
    def __recoverable(self, error, url, request, counter=1):
        self.logger.debug(f"Connection exception handler: {error}")
        if not (counter < self.retry_max):
            self.logger.debug("Max retries of '{}' reached".format(self.retry_max))
            raise OasisException(f'Failed to recover from error: {error}', error)

        if isinstance(error, (ConnectionError, ReadTimeout)):
            self.logger.debug(f"Recoverable error [{error}] from {request} {url}")
            self.logger.debug(f"Backoff timer: {self.retry_delay * counter}")

            # Reset HTTPAdapter & clear connection pool
            self.adapters.clear()
            self.mount(self.url_base, HTTPAdapter(max_retries=self.retry_max))
            time.sleep(self.retry_delay * counter)
            return True

        elif isinstance(error, HTTPError):
            http_err_code = error.response.status_code
            self.logger.debug(f"Recoverable error [{error}] from {request} {url}")

            if http_err_code in [502, 503, 504]:
                error = "HTTP {}".format(http_err_code)
                return True
            elif http_err_code in [401, 403]:
                if self.tkn_refresh is not None:
                    self.logger.debug("requesting refresh token")
                    self._refresh_token()
                    return True
        return False

    # @oasis_log
    def health_check(self):
        """
        Checks the health of the server.

        """
        try:
            url = urljoin(self.url_base, 'healthcheck/')
            r = super(APISession, self).get(url)
            r.raise_for_status()
            return r
        except (TypeError, AttributeError, BytesWarning, HTTPError, ConnectionError, ReadTimeout) as e:
            err_msg = 'Health check failed: Unable to connect to {}'.format(self.url_base)
            raise OasisException(err_msg, e)

    def upload(self, url, filepath, content_type, **kwargs):
        counter = 0
        while True:
            counter += 1
            try:
                abs_fp = os.path.realpath(os.path.expanduser(filepath))
                m = MultipartEncoder(fields={'file': (os.path.basename(filepath), open(abs_fp, 'rb'), content_type)})
                r = super(APISession, self).post(url, data=m, headers={'Content-Type': m.content_type}, timeout=self.timeout, **kwargs)
                r.raise_for_status()
                time.sleep(self.request_interval)
            except (HTTPError, ConnectionError, ReadTimeout) as e:
                if self.__recoverable(e, url, 'GET', counter):
                    continue
                else:
                    self.logger.debug(f'Unrecoverable error: {e}')
                    raise e
            return r

    def upload_byte(self, url, file_bytes, filename, content_type, **kwargs):
        counter = 0
        while True:
            counter += 1
            try:
                m = MultipartEncoder(fields={'file': (filename, file_bytes, content_type)})
                r = super(APISession, self).post(url, data=m,
                                                 headers={'Content-Type': m.content_type},
                                                 timeout=self.timeout,
                                                 **kwargs)
                r.raise_for_status()
                time.sleep(self.request_interval)
            except (HTTPError, ConnectionError, ReadTimeout) as e:
                if self.__recoverable(e, url, 'GET', counter):
                    continue
                else:
                    self.logger.debug(f'Unrecoverable error: {e}')
                    raise e
            return r

    def get(self, url, **kwargs):
        counter = 0
        while True:
            counter += 1
            try:
                r = super(APISession, self).get(url, timeout=self.timeout, **kwargs)
                r.raise_for_status()
                time.sleep(self.request_interval)
            except (HTTPError, ConnectionError, ReadTimeout) as e:
                if self.__recoverable(e, url, 'GET', counter):
                    continue
                else:
                    self.logger.debug(f'Unrecoverable error: {e}')
                    raise e
            return r

    def post(self, url, **kwargs):
        counter = 0
        while True:
            counter += 1
            try:
                r = super(APISession, self).post(url, timeout=self.timeout, **kwargs)
                r.raise_for_status()
                time.sleep(self.request_interval)
            except (HTTPError, ConnectionError, ReadTimeout) as e:
                if self.__recoverable(e, url, 'POST', counter):
                    continue
                else:
                    self.logger.debug(f'Unrecoverable error: {e}')
                    raise e
            return r

    def delete(self, url, **kwargs):
        counter = 0
        while True:
            counter += 1
            try:
                r = super(APISession, self).delete(url, timeout=self.timeout, **kwargs)
                r.raise_for_status()
                time.sleep(self.request_interval)
            except (HTTPError, ConnectionError, ReadTimeout) as e:
                if self.__recoverable(e, url, 'DELETE', counter):
                    continue
                else:
                    self.logger.debug(f'Unrecoverable error: {e}')
                    raise e
            return r

    def put(self, url, **kwargs):
        counter = 0
        while True:
            counter += 1
            try:
                r = super(APISession, self).put(url, timeout=self.timeout, **kwargs)
                r.raise_for_status()
                time.sleep(self.request_interval)
            except (HTTPError, ConnectionError, ReadTimeout) as e:
                if self.__recoverable(e, url, 'PUT', counter):
                    continue
                else:
                    self.logger.debug(f'Unrecoverable error: {e}')
                    raise e
            return r



#!/usr/bin/env python3

import re

import numpy as np
import pandas as pd
import os
import json
import argparse

parser = argparse.ArgumentParser()

parser.add_argument("-dir", "--subdirectory", help="enter subdirectory name")

params = parser.parse_args()

subdir = params.subdirectory


os.chdir(subdir)

location_file = pd.read_csv('location.csv', dtype=str, keep_default_na=False)
account_file = pd.read_csv('account.csv', dtype=str, keep_default_na=False)
if os.path.isfile('ri_info.csv') and os.path.isfile('ri_scope.csv'):
    ri_info_file = pd.read_csv('ri_info.csv', dtype=str, keep_default_na=False)
    ri_scope_file = pd.read_csv('ri_scope.csv', dtype=str, keep_default_na=False)
else:
    ri_scope_file = None

split_location = location_file.groupby('FlexiLocUnit')
split_account = account_file.groupby('FlexiAccUnit')

newpath = 'units'
if not os.path.exists(newpath):
    os.makedirs(newpath)

cwd = os.getcwd()

# loop through the groups and save to directories based on unique values
for name, group in split_location:
    sub_dir = os.path.join(newpath, name)
    if not os.path.exists(sub_dir):
        os.mkdir(sub_dir)
    group = group.drop(['FlexiLocUnit'], axis=1)

    group.to_csv(sub_dir + "/location.csv", index=0)

    if ri_scope_file is not None:
        ri_scope = ri_scope_file[(ri_scope_file['PortNumber'].isin(np.unique(group['PortNumber'])))
                                 & (ri_scope_file['AccNumber'].isin(list(np.unique(group['AccNumber'])) + ['']))]
        ri_info = ri_info_file[ri_info_file['ReinsNumber'].isin(np.unique(ri_scope['ReinsNumber']))]
        if not ri_scope.empty:
            ri_scope.to_csv(sub_dir + "/ri_scope.csv", index=0)
            ri_info.to_csv(sub_dir + "/ri_info.csv", index=0)


for name, group in split_account:
    sub_dir = os.path.join(newpath, name)
    if not os.path.exists(sub_dir):
        os.mkdir(sub_dir)
    group = group.drop(['FlexiAccUnit'], axis=1)
    group.to_csv(sub_dir + "/account.csv", index=0)

names = sorted([str(item[0]) for item in split_location])

# Function to sort fm string in Ascedning order


def ascedning(text):
    return int(text) if text.isdigit() else text


def natural_keys(text):

    return [ascedning(c) for c in re.split(r'(\d+)', text)]


names.sort(key=natural_keys)
# print(names)

units_dir = os.path.join(cwd, 'units')

if not os.path.exists(units_dir):
    os.mkdir(units_dir)

with open(os.path.join(units_dir, 'units.txt'), "w") as txt_file:
    names, groups = map(list, zip(*split_location))
    names.sort(key=natural_keys)
    for name in names:
        txt_file.write(str(name) + '\n')



#!/usr/bin/env python3

import argparse
import os
import io
import subprocess
import sys
from tabulate import tabulate


def parse_args():
    desc = (
        'Performance testing script for OasisLMF input file generation'
        'This script expects a set of nested sub directories each containing'
        'acc.csv, loc.csv, keys.csv'
    )
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument('-t', '--time-threshold', default=300, type=int, help='Maximum time for each file generation test')
    parser.add_argument('-d', '--test-directory', default='.', type=str, help='File path of the test data directory')
    parser.add_argument('-o', '--output-directory', default='/tmp/oasis-files', type=str, help='Filepath to generate oasisfiles in')
    parser.add_argument('-l', '--log-output', default='/var/report/oasisfiles_benchmark.log', type=str, help='Log file path')
    parser.add_argument('-a', '--extra-oasislmf-args', default='', type=str, help='Addtional Aguments to run Oasislmf with')
    return vars(parser.parse_args())


def run_command(cmd_str):
    resp = subprocess.run(cmd_str.split(), stdout=subprocess.PIPE, stderr=subprocess.STDOUT, check=False)
    stdout = resp.stdout.decode('utf-8').strip()
    print(stdout)
    resp.check_returncode()
    return stdout


def pasrse_gen_output(stdout_string):
    runtime_list = [l for l in stdout_string.split('\n') if 'COMPLETED' in l]
    t_breakdown = dict()
    total = runtime_list.pop().rsplit(' ').pop()
    t_breakdown['total'] = float(total[:-1])

    for l in runtime_list:
        line = l.split(' ')
        func = line[-3]
        time = line[-1]
        t_breakdown[func] = float(time[:-1])
    return t_breakdown


def tabulate_data(test_results, output_fp=None):
    input_sizes = sorted(list(test_results.keys()))
    time_values = dict()
    func_names = test_results[input_sizes[0]].keys()

    for f in func_names:
        name = f.split('.')[-1:].pop()
        time_values[name] = list()

    for n in input_sizes:
        for f in func_names:
            name = f.split('.')[-1:].pop()
            time_values[name].append(test_results[n][f])

    timing_tbl = tabulate(
        [[k] + time_values[k] for k in time_values],
        headers=['portfolio size'] + input_sizes,
        tablefmt="rst")

    # if set write to test summary table to log file
    if output_fp:
        log_path = os.path.abspath(output_fp)
        log_dir = os.path.dirname(log_path)
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)

        with io.open(log_path, 'w') as log:
            log.write(timing_tbl)

    print(timing_tbl)


def run_tests(test_dir, run_dir, log_fp, oasis_args, threshold=None):
    '''
    Output of each run entry in `results`

    In [3]: example_run
    Out[3]:
    {'total': 88.63,
     'oasislmf.manager.__init__': 0.0,
     'oasislmf.model_preparation.gul_inputs.get_gul_input_items': 16.05,
     'oasislmf.model_preparation.gul_inputs.write_items_file': 3.84,
     'oasislmf.model_preparation.gul_inputs.write_coverages_file': 1.88,
     'oasislmf.model_preparation.gul_inputs.write_gul_input_files': 5.94,
     'oasislmf.model_preparation.summaries.get_summary_mapping': 0.8,
     'oasislmf.model_preparation.summaries.write_mapping_file': 6.77,
     'oasislmf.model_preparation.il_inputs.get_il_input_items': 30.42,
     'oasislmf.model_preparation.il_inputs.write_fm_policytc_file': 8.49,
     'oasislmf.model_preparation.il_inputs.write_fm_profile_file': 1.59,
     'oasislmf.model_preparation.il_inputs.write_fm_programme_file': 7.52,
     'oasislmf.model_preparation.il_inputs.write_fm_xref_file': 2.98,
     'oasislmf.model_preparation.il_inputs.write_il_input_files': 21.44}
    '''
    sub_dirs = next(os.walk(test_dir))[1]
    test_data = dict()
    results = dict()

    for d in sub_dirs:
        loc_fp = os.path.join(test_dir, d, 'loc.csv')
        acc_fp = os.path.join(test_dir, d, 'acc.csv')
        keys_fp = os.path.join(test_dir, d, 'keys.csv')

        n_sample = sum(1 for line in open(loc_fp)) - 1
        cmd_str = f'oasislmf model generate-oasis-files -x {loc_fp} -y {acc_fp} -z {keys_fp} --oasis-files-dir {run_dir} {oasis_args} --verbose'
        test_data[n_sample] = cmd_str

    for t in sorted(test_data.keys()):
        print('Running: ')
        print(f"cmd = {test_data[t]}")
        print(f'size = {t}')
        print(f't_max = {threshold}')
        stdout = run_command(test_data[t])
        run = pasrse_gen_output(stdout)
        results[t] = run
        print(f"t_total = {run['total']}\n")

        # If given check that threshold isn't exceeded
        if threshold:
            if run['total'] > threshold:
                print('FAILED\n')
                tabulate_data(results, log_fp)
                sys.exit(1)
            else:
                print('PASSED\n')

    tabulate_data(results, log_fp)
    return results


if __name__ == "__main__":
    args = parse_args()
    run_tests(args['test_directory'],
              args['output_directory'],
              args['log_output'],
              args['extra_oasislmf_args'],
              args['time_threshold'])
